#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""State Module."""

from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, TypedDict, TypeVar, cast

from molecule import util


if TYPE_CHECKING:
    from molecule.config import Config
VALID_KEYS = [
    "created",
    "converged",
    "driver",
    "prepared",
    "run_uuid",
    "is_parallel",
    "molecule_yml_date_modified",
]
F = TypeVar("F", bound=Callable[..., None])


class InvalidState(Exception):  # noqa: N818
    """Exception class raised when an error occurs in :class:`.State`."""


class StateData(TypedDict):
    """Valid state values and types.

    Attributes:
        converged: Has scenario converged.
        created: Has scenario been created:
        driver: Driver for scenario.
        prepared: Has scenario prepare run.
        molecule_yml_date_modified: Modified date of molecule.yml file.
        run_uuid: UUID of active run.
        is_parallel: Is this run parallel.
    """

    converged: bool
    created: bool
    driver: str | None
    prepared: bool
    molecule_yml_date_modified: float | None
    run_uuid: str
    is_parallel: bool


def marshal(func: F) -> F:
    """Decorator to immediately write state to file after call finishes.

    Args:
        func: Function to decorate.

    Returns:
        Decorated function.
    """

    def wrapper(self: State, *args: object, **kwargs: object) -> None:
        func(self, *args, **kwargs)
        self._write_state_file()

    return cast("F", wrapper)


class State:
    """A class which manages the state file.

    Intended to be used as a singleton throughout a given Molecule config.
    The initial state is serialized to disk if the file does not exist,
    otherwise is deserialized from the existing state file.  Changes made to
    the object are immediately serialized.

    State is not a top level option in Molecule's config.  It's purpose is for
    bookkeeping, and each :class:`.Config` object has a reference to a State_
    object.

    !!! note

        Currently, it's use is significantly smaller than it was in v1 of
        Molecule.
    """

    def __init__(self, config: Config) -> None:
        """Initialize a new state class and returns None.

        Args:
            config: An instance of a Molecule config.
        """
        self._config = config
        self._state_file = self._get_state_file()
        self._data = self._get_data()
        self._write_state_file()

    @property
    def state_file(self) -> str:
        """State file path.

        Returns:
            Path to the state file.
        """
        return str(self._state_file)

    @property
    def converged(self) -> bool:
        """Is run converged.

        Returns:
            Whether the run has converged.
        """
        return self._data["converged"]

    @property
    def created(self) -> bool:
        """Has scenario been created.

        Returns:
            Whether scenario has been created.
        """
        return self._data["created"]

    @property
    def driver(self) -> str | None:
        """Driver for scenario.

        Returns:
            Name of the driver for the scenario.
        """
        return self._data["driver"]

    @property
    def prepared(self) -> bool:
        """Has scenario prepare run.

        Returns:
            Whether scenario prepare has run.
        """
        return self._data["prepared"]

    @property
    def run_uuid(self) -> str:
        """Scenario run UUID.

        Returns:
            The UUID for this scenario run.
        """
        return self._data["run_uuid"]

    @property
    def is_parallel(self) -> bool:
        """Is molecule in parallel mode.

        Returns:
            Whether Molecule is in parallel mode.
        """
        return self._data["is_parallel"]

    @property
    def molecule_yml_date_modified(self) -> float | None:
        """The modified date of molecule.yml.

        Returns:
            The timestamp of the last modification date of molecule.yml.
        """
        return self._data["molecule_yml_date_modified"]

    @marshal
    def reset(self) -> None:
        """Reset state data."""
        self._data = self._default_data()

    @marshal
    def change_state(self, key: str, value: object) -> None:
        """Change the state of the instance data with the given `key` and the provided ``value`.

        Wrapping with a decorator is probably not necessary.

        Args:
            key: A ``str`` containing the key to update
            value: A value to change the ``key`` to

        Raises:
            InvalidState: if an invalid key is requested.
        """
        if key not in VALID_KEYS:
            raise InvalidState
        self._data[key] = value  # type: ignore[literal-required]

    def _get_data(self) -> StateData:
        if self._state_file.is_file():
            return self._load_file()
        return self._default_data()

    def _default_data(self) -> StateData:
        return {
            "converged": False,
            "created": False,
            "driver": None,
            "prepared": False,
            "molecule_yml_date_modified": None,
            "run_uuid": self._config._run_uuid,  # noqa: SLF001
            "is_parallel": self._config.is_parallel,
        }

    def _load_file(self) -> StateData:
        return cast("StateData", util.safe_load_file(self._state_file))

    def _write_state_file(self) -> None:
        util.write_file(self.state_file, util.safe_dump(self._data))

    def _get_state_file(self) -> Path:
        return Path(self._config.scenario.ephemeral_directory) / "state.yml"


class SharedState(State):
    """A class which manages the state file.

    This is exactly like the normal State object but it's backed by the shared ephemeral directory,
    so state is maintained between scenarios.
    """

    def _get_state_file(self) -> Path:
        return Path(self._config.scenario.shared_ephemeral_directory) / "state.yml"



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os

import pytest

from molecule import config, util
from molecule.provisioner import ansible_playbooks
from tests.unit.conftest import os_split  # pylint:disable=C0411


@pytest.fixture
def _provisioner_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"provisioner": {"name": "ansible", "options": {}, "config_options": {}}}


@pytest.fixture
def _instance(_provisioner_section_data, config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return ansible_playbooks.AnsiblePlaybooks(config_instance)


def test_cleanup_property_is_optional(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._config.provisioner.playbooks.cleanup is None


@pytest.mark.skip(reason="create not running for delegated")
def test_create_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = os.path.join(_instance._get_playbook_directory(), "default", "create.yml")  # noqa: PTH118

    assert x == _instance._config.provisioner.playbooks.create


def test_converge_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = os.path.join(_instance._config.scenario.directory, "converge.yml")  # noqa: PTH118

    assert x == _instance._config.provisioner.playbooks.converge


@pytest.mark.skip(reason="destroy not running for delegated")
def test_destroy_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = os.path.join(_instance._get_playbook_directory(), "default", "destroy.yml")  # noqa: PTH118

    assert x == _instance._config.provisioner.playbooks.destroy


def test_prepare_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._config.provisioner.playbooks.prepare is None


def test_side_effect_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._config.provisioner.playbooks.side_effect is None


def test_verify_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._config.provisioner.playbooks.verify is None


def test_get_playbook_directory(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    result = _instance._get_playbook_directory()
    parts = os_split(result)
    x = ("molecule", "provisioner", "ansible", "playbooks")

    assert x == parts[-4:]


def test_get_playbook(tmpdir, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = os.path.join(_instance._config.scenario.directory, "create.yml")  # noqa: PTH118
    util.write_file(x, "")

    assert x == _instance._get_playbook("create")


@pytest.mark.skip(reason="create not running for delegated")
def test_get_playbook_returns_bundled_driver_playbook_when_local_not_found(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    tmpdir,
    _instance,  # noqa: PT019
):
    x = os.path.join(_instance._get_playbook_directory(), "default", "create.yml")  # noqa: PTH118

    assert x == _instance._get_playbook("create")


@pytest.fixture
def _provisioner_driver_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "provisioner": {
            "name": "ansible",
            "playbooks": {
                "create": "create.yml",
            },
        },
    }


@pytest.fixture
def _provisioner_driver_playbook_key_missing_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "provisioner": {
            "name": "ansible",
            "playbooks": {
                "side_effect": "side_effect.yml",
            },
        },
    }


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_driver_playbook_key_missing_section_data"],  # noqa: PT007
    indirect=True,
)
def test_get_ansible_playbook_with_driver_key_when_playbook_key_missing(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    tmpdir,
    _instance,  # noqa: PT019
):
    x = os.path.join(_instance._config.scenario.directory, "side_effect.yml")  # noqa: PTH118
    util.write_file(x, "")

    assert x == _instance._get_playbook("side_effect")


def test_get_bundled_driver_playbook(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    result = _instance._get_bundled_driver_playbook("create")
    parts = os_split(result)
    x = ("molecule", "driver", "playbooks", "create.yml")

    assert x == parts[-4:]



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Lint Command Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molecule.api import drivers
from molecule.click_cfg import click_command_ex, options
from molecule.command import base


if TYPE_CHECKING:
    import click

    from molecule.types import CommandArgs


@click_command_ex()
@options(["scenario_name_single_with_default"])
def reset(ctx: click.Context) -> None:  # pragma: no cover
    """Reset molecule temporary folders.

    Args:
        ctx: Click context object holding commandline arguments.
    """
    args = ctx.obj.get("args")
    subcommand = base._get_subcommand(__name__)  # noqa: SLF001
    command_args: CommandArgs = {"subcommand": subcommand}

    scenario_name = ctx.params["scenario_name"]

    base.execute_cmdline_scenarios([scenario_name], args, command_args)
    for driver in drivers().values():
        driver.reset()



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from molecule.command import login
from molecule.exceptions import MoleculeError


if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from molecule import config


@pytest.fixture
def _instance(config_instance: config.Config) -> login.Login:
    config_instance.state.change_state("created", value=True)

    return login.Login(config_instance)


def test_login_execute(  # noqa: D103
    mocker: MockerFixture,
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "instance-1"}
    m = mocker.patch("molecule.command.login.Login._get_login")
    _instance.execute()

    m.assert_called_once_with("instance-1")


@pytest.mark.parametrize(
    "config_instance",
    ["command_driver_delegated_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_login_execute_instance_creation(  # noqa: D103
    mocker: MockerFixture,
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "instance-1"}
    _instance._config.state.change_state("created", value=False)

    mocker.patch("molecule.command.login.Login._get_login")
    patched_execute_subcommand = mocker.patch("molecule.command.base.execute_subcommand")
    patched_execute_subcommand.side_effect = lambda _config, _: _config.state.change_state(
        key="created",
        value=True,
    )
    _instance.execute()

    patched_execute_subcommand.assert_called_once_with(_instance._config, "create")
    assert _instance._config.state.created


def test_get_hostname_does_not_match(  # noqa: D103
    caplog: pytest.LogCaptureFixture,
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "invalid"}
    hosts = ["instance-1"]
    with pytest.raises(MoleculeError) as e:
        _instance._get_hostname(hosts)

    assert e.value.code == 1

    msg = "There are no hosts that match 'invalid'.  You can only login to valid hosts."
    assert msg in caplog.text


def test_get_hostname_exact_match_with_one_host(  # noqa: D103
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "instance-1"}
    hosts = ["instance-1"]

    assert _instance._get_hostname(hosts) == "instance-1"


def test_get_hostname_partial_match_with_one_host(  # noqa: D103
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "inst"}
    hosts = ["instance-1"]

    assert _instance._get_hostname(hosts) == "instance-1"


def test_get_hostname_exact_match_with_multiple_hosts(  # noqa: D103
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "instance-1"}
    hosts = ["instance-1", "instance-2"]

    assert _instance._get_hostname(hosts) == "instance-1"


def test_get_hostname_partial_match_with_multiple_hosts(  # noqa: D103
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "foo"}
    hosts = ["foo", "fooo"]

    assert _instance._get_hostname(hosts) == "foo"


def test_get_hostname_partial_match_with_multiple_hosts_raises(  # noqa: D103
    caplog: pytest.LogCaptureFixture,
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {"host": "inst"}
    hosts = ["instance-1", "instance-2"]
    with pytest.raises(MoleculeError) as e:
        _instance._get_hostname(hosts)

    assert e.value.code == 1

    msg = (
        "There are 2 hosts that match 'inst'. "
        "You can only login to one at a time.\n\n"
        "Available hosts:\n"
        "instance-1\n"
        "instance-2"
    )
    assert msg in caplog.text


def test_get_hostname_no_host_flag_specified_on_cli(  # noqa: D103
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {}
    hosts = ["instance-1"]
    _instance._get_hostname(hosts)

    assert _instance._get_hostname(hosts) == "instance-1"


def test_get_hostname_no_host_flag_specified_on_cli_with_multiple_hosts_raises(  # noqa: D103
    caplog: pytest.LogCaptureFixture,
    _instance: login.Login,  # noqa: PT019
) -> None:
    _instance._config.command_args = {}
    hosts = ["instance-1", "instance-2"]
    with pytest.raises(MoleculeError) as e:
        _instance._get_hostname(hosts)

    assert e.value.code == 1

    msg = (
        "There are 2 running hosts. Please specify "
        "which with --host.\n\n"
        "Available hosts:\n"
        "instance-1\n"
        "instance-2"
    )
    assert msg in caplog.text



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Idempotence Command Module."""

from __future__ import annotations

import re

from typing import TYPE_CHECKING

from molecule.click_cfg import click_command_ex, common_options
from molecule.command import base
from molecule.exceptions import ScenarioFailureError
from molecule.text import strip_ansi_escape


if TYPE_CHECKING:
    import click

    from molecule.types import CommandArgs


class Idempotence(base.Base):
    """Runs the converge step a second time.

    If no tasks will be marked as changed \
    the scenario will be considered idempotent.
    """

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute the actions necessary to perform a `molecule idempotence`.

        Args:
            action_args: Arguments for this command. Unused.

        Raises:
            ScenarioFailureError: when idempotence fails to run.
        """
        if not self._config.state.converged:
            msg = "Instances not converged.  Please converge instances first."
            raise ScenarioFailureError(message=msg)

        if self._config.provisioner:
            output = self._config.provisioner.converge()

            idempotent = self._is_idempotent(output)
            if idempotent:
                msg = "Idempotence completed successfully."
                self._log.info(msg)
            else:
                details = "\n".join(self._non_idempotent_tasks(output))
                msg = f"Idempotence test failed because of the following tasks:\n{details}"
                raise ScenarioFailureError(message=msg)

    def _is_idempotent(self, output: str) -> bool:
        """Parse the output of the provisioning for changed and returns a bool.

        Args:
            output: A string containing the output of the ansible run.

        Returns:
            bool: True if the output is idempotent, False otherwise.
        """
        # Remove blank lines to make regex matches easier
        output = re.sub(r"\n\s*\n*", "\n", output)

        # Look for any non-zero changed lines
        changed = re.search(r"(changed=[1-9][0-9]*)", output)

        return not bool(changed)

    def _non_idempotent_tasks(self, output: str) -> list[str]:
        """Parse the output to identify the non idempotent tasks.

        Args:
            output: A string containing the output of the ansible run.

        Returns:
            list: A list containing the names of the non idempotent tasks.

        """
        # Remove blank lines to make regex matches easier.
        output = re.sub(r"\n\s*\n*", "\n", output)

        # Remove ansi escape sequences.
        output = strip_ansi_escape(output)

        # Split the output into a list and go through it.
        output_lines = output.split("\n")
        res = []
        task_line = ""
        for _, line in enumerate(output_lines):
            if line.startswith("TASK"):
                task_line = line
            elif line.startswith("changed"):
                host_name = re.search(r"\[(.*)\]", line).groups()[0]  # type: ignore[union-attr]
                task_name = re.search(r"\[(.*)\]", task_line).groups()[0]  # type: ignore[union-attr]
                res.append(f"* [{host_name}] => {task_name}")

        return res


@click_command_ex()
@common_options("ansible_args")
def idempotence(ctx: click.Context) -> None:  # pragma: no cover
    """Use the provisioner to configure the instances.

    After parse the output to determine idempotence.

    Args:
        ctx: Click context object holding commandline arguments.
    """
    args = ctx.obj.get("args")
    subcommand = base._get_subcommand(__name__)  # noqa: SLF001
    command_args: CommandArgs = {
        "report": ctx.params["report"],
        "shared_inventory": ctx.params["shared_inventory"],
        "shared_state": ctx.params["shared_state"],
        "subcommand": subcommand,
    }

    __all = ctx.params["all"]
    ansible_args = ctx.params["ansible_args"]
    exclude = ctx.params["exclude"]
    scenario_name = ctx.params["scenario_name"]

    if __all:
        scenario_name = None

    base.execute_cmdline_scenarios(scenario_name, args, command_args, ansible_args, exclude)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import pytest

from molecule.model import schema_v3


@pytest.fixture
def _model_driver_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "driver": {
            "name": "default",
            "provider": {"name": None},
            "options": {"managed": True},
            "ssh_connection_options": ["foo", "bar"],
            "safe_files": ["foo", "bar"],
        },
    }


@pytest.mark.parametrize("config", ["_model_driver_section_data"], indirect=True)  # noqa: PT007
def test_driver(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert not schema_v3.validate(config)


@pytest.fixture
def _model_driver_errors_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "driver": {
            "name": 0,
        },
    }


@pytest.fixture
def _model_driver_errors_section_data_no_prefix():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "driver": {
            "name": "random_name",
        },
    }


@pytest.mark.parametrize(
    "config",
    [  # noqa: PT007
        "_model_driver_errors_section_data",
        "_model_driver_errors_section_data_no_prefix",
    ],
    indirect=True,
)
def test_driver_has_errors(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    base_error_msg = "is not one of ['azure', 'ec2', 'delegated', 'docker', 'containers', 'openstack', 'podman', 'vagrant', 'digitalocean', 'gce', 'libvirt', 'lxd', 'molecule-*', 'molecule_*', 'custom-*', 'custom_*']"

    driver_name = str(config["driver"]["name"])
    if isinstance(config["driver"]["name"], str):
        # add single quotes for string
        driver_name = f"'{config['driver']['name']}'"

    error_msg = [f"{driver_name} {base_error_msg}"]
    assert error_msg == schema_v3.validate(config)


@pytest.fixture
def _model_driver_provider_name_nullable_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"provider": {"name": None}}}


@pytest.mark.parametrize(
    "config",
    ["_model_driver_provider_name_nullable_section_data"],  # noqa: PT007
    indirect=True,
)
def test_driver_provider_name_nullable(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert not schema_v3.validate(config)


@pytest.fixture
def _model_driver_allows_delegated_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"name": "default"}}


@pytest.fixture
def _model_driver_allows_molecule_section_data1():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"name": "molecule-test_driver.name"}}


@pytest.fixture
def _model_driver_allows_molecule_section_data2():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"name": "molecule_test_driver.name"}}


@pytest.fixture
def _model_driver_allows_custom_section_data1():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"name": "custom-test_driver.name"}}


@pytest.fixture
def _model_driver_allows_custom_section_data2():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"name": "custom_test_driver.name"}}


###
@pytest.mark.parametrize(
    "config",
    [  # noqa: PT007
        ("_model_driver_allows_delegated_section_data"),
        ("_model_driver_allows_molecule_section_data1"),
        ("_model_driver_allows_molecule_section_data2"),
        ("_model_driver_allows_custom_section_data2"),
        ("_model_driver_allows_custom_section_data1"),
    ],
    indirect=True,
)
def test_driver_allows_name(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert not schema_v3.validate(config)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.

#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Ansible Provisioner Module."""

from __future__ import annotations

import collections
import copy
import os
import shutil

from pathlib import Path
from typing import TYPE_CHECKING

from ansible_compat.ports import cached_property

from molecule import logger, util
from molecule.exceptions import MoleculeError
from molecule.provisioner import ansible_playbook, ansible_playbooks, base


if TYPE_CHECKING:
    from typing import Any

    from molecule.types import Options

    Vivify = collections.defaultdict[str, Any | "Vivify"]


class Ansible(base.Base):
    """`Ansible` is the default provisioner.  No other provisioner will be supported.

    Molecule's provisioner manages the instances lifecycle.  However, the user
    must provide the create, destroy, and converge playbooks.  Molecule's
    ``init`` subcommand will provide the necessary files for convenience.

    Molecule will skip tasks which are tagged with either `molecule-notest` or
    `notest`. With the tag `molecule-idempotence-notest` tasks are only
    skipped during the idempotence action step.

    !!! warning

        Reserve the create and destroy playbooks for provisioning.  Do not
        attempt to gather facts or perform operations on the provisioned nodes
        inside these playbooks.  Due to the gymnastics necessary to sync state
        between Ansible and Molecule, it is best to perform these tasks in the
        prepare or converge playbooks.

        It is the developers responsibility to properly map the modules' fact
        data into the instance_conf_dict fact in the create playbook.  This
        allows Molecule to properly configure Ansible inventory.

    Additional options can be passed to ``ansible-playbook`` through the options
    dict.  Any option set in this section will override the defaults.

    !!! note

        Options do not affect the create and destroy actions.

    !!! note

        Molecule will remove any options matching `^[v]+$`, and pass ``-vvv``
        to the underlying ``ansible-playbook`` command when executing
        `molecule --debug`.

    Molecule will silence log output, unless invoked with the ``--debug`` flag.
    However, this results in quite a bit of output.  To enable Ansible log
    output, add the following to the ``provisioner`` section of ``molecule.yml``.

    ``` yaml
        provisioner:
          name: ansible
          log: True
    ```

    The create/destroy playbooks for Docker and Podman are bundled with
    Molecule.  These playbooks have a clean API from `molecule.yml`, and
    are the most commonly used.  The bundled playbooks can still be overridden.

    The playbook loading order is:

    1. provisioner.playbooks.$driver_name.$action
    2. provisioner.playbooks.$action
    3. bundled_playbook.$driver_name.$action

    ``` yaml
        provisioner:
          name: ansible
          options:
            vvv: True
          playbooks:
            create: create.yml
            converge: converge.yml
            destroy: destroy.yml
    ```

    Share playbooks between roles.

    ``` yaml
        provisioner:
          name: ansible
          playbooks:
            create: ../default/create.yml
            destroy: ../default/destroy.yml
            converge: converge.yml
    ```

    Multiple driver playbooks.  In some situations a developer may choose to
    test the same role against different backends.  Molecule will choose driver
    specific create/destroy playbooks, if the determined driver has a key in
    the playbooks section of the provisioner's dict.

    !!! note

        If the determined driver has a key in the playbooks dict, Molecule will
        use this dict to resolve all provisioning playbooks (create/destroy).

    ``` yaml
        provisioner:
          name: ansible
          playbooks:
            docker:
              create: create.yml
              destroy: destroy.yml
            create: create.yml
            destroy: destroy.yml
            converge: converge.yml
    ```

    !!! note

        Paths in this section are converted to absolute paths, where the
        relative parent is the $scenario_directory.

    The side effect playbook executes actions which produce side effects to the
    instances(s).  Intended to test HA failover scenarios or the like.  It is
    not enabled by default.  Add the following to the provisioner's ``playbooks``
    section to enable.

    ``` yaml
        provisioner:
          name: ansible
          playbooks:
            side_effect: side_effect.yml
    ```

    !!! note

        This feature should be considered experimental.

    The prepare playbook executes actions which bring the system to a given
    state prior to converge.  It is executed after create, and only once for
    the duration of the instances life.

    This can be used to bring instances into a particular state, prior to
    testing.

    ``` yaml
        provisioner:
          name: ansible
          playbooks:
            prepare: prepare.yml
    ```

    The cleanup playbook is for cleaning up test infrastructure that may not
    be present on the instance that will be destroyed. The primary use-case
    is for "cleaning up" changes that were made outside of Molecule's test
    environment. For example, remote database connections or user accounts.
    Intended to be used in conjunction with `prepare` to modify external
    resources when required.

    The cleanup step is executed directly before every destroy step. Just like
    the destroy step, it will be run twice. An initial clean before converge
    and then a clean before the last destroy step. This means that the cleanup
    playbook must handle failures to cleanup resources which have not
    been created yet.

    Add the following to the provisioner's `playbooks` section
    to enable.

    ``` yaml
        provisioner:
          name: ansible
          playbooks:
            cleanup: cleanup.yml
    ```

    !!! note

        This feature should be considered experimental.

    Environment variables can be passed to the provisioner.  Variables in this
    section which match the names above will be appended to the above defaults,
    and converted to absolute paths, where the relative parent is the
    $scenario_directory.

    !!! note

        Paths in this section are converted to absolute paths, where the
        relative parent is the $scenario_directory.

    ``` yaml
        provisioner:
          name: ansible
          env:
            FOO: bar
    ```

    Modifying ansible.cfg.

    ``` yaml
        provisioner:
          name: ansible
          config_options:
            defaults:
              fact_caching: jsonfile
            ssh_connection:
              scp_if_ssh: True
    ```

    !!! note

        The following keys are disallowed to prevent Molecule from
        improperly functioning.  They can be specified through the
        provisioner's env setting described above, with the exception
        of the `privilege_escalation`.

    ``` yaml
        provisioner:
          name: ansible
          config_options:
            defaults:
              library: /path/to/library
              filter_plugins: /path/to/filter_plugins
            privilege_escalation: {}
    ```

    Roles which require host/groups to have certain variables set.  Molecule
    uses the same `variables defined in a playbook`_ syntax as `Ansible`_.

    ``` yaml
        provisioner:
          name: ansible
          inventory:
            group_vars:
              all:
                bar: foo
              group1:
                foo: bar
              group2:
                foo: bar
                baz:
                  qux: zzyzx
            host_vars:
              group1-01:
                foo: baz
    ```

    Molecule automatically generates the inventory based on the hosts defined
    under `Platforms`_. Using the ``hosts`` key allows to add extra hosts to
    the inventory that are not managed by Molecule.

    A typical use case is if you want to access some variables from another
    host in the inventory (using hostvars) without creating it.

    !!! note

        The content of ``hosts`` should follow the YAML based inventory syntax:
        start with the ``all`` group and have hosts/vars/children entries.

    ``` yaml
        provisioner:
          name: ansible
          inventory:
            hosts:
              all:
                hosts:
                  extra_host:
                    foo: hello
    ```

    !!! note

        The extra hosts added to the inventory using this key won't be
        created/destroyed by Molecule. It is the developers responsibility
        to target the proper hosts in the playbook. Only the hosts defined
        under `Platforms`_ should be targeted instead of ``all``.


    An alternative to the above is symlinking.  Molecule creates symlinks to
    the specified directory in the inventory directory. This allows ansible to
    converge utilizing its built in host/group_vars resolution. These two
    forms of inventory management are mutually exclusive.

    Like above, it is possible to pass an additional inventory file
    (or even dynamic inventory script), using the ``hosts`` key. `Ansible`_ will
    automatically merge this inventory with the one generated by molecule.
    This can be useful if you want to define extra hosts that are not managed
    by Molecule.

    !!! note

        Again, it is the developers responsibility to target the proper hosts
        in the playbook. Only the hosts defined under
        `Platforms`_ should be targeted instead of ``all``.

    !!! note

        The source directory linking is relative to the scenario's
        directory.

        The only valid keys are ``hosts``, ``group_vars`` and ``host_vars``.  Molecule's
        schema validator will enforce this.


    ``` yaml
        provisioner:
          name: ansible
          inventory:
            links:
              hosts: ../../../inventory/hosts
              group_vars: ../../../inventory/group_vars/
              host_vars: ../../../inventory/host_vars/
    ```

    !!! note

        You can use either `hosts`/`group_vars`/`host_vars` sections of inventory OR `links`.
        If you use both, links will win.

    ``` yaml
        provisioner:
          name: ansible
          hosts:
            all:
              hosts:
                ignored:
                   important: this host is ignored
          inventory:
            links:
              hosts: ../../../inventory/hosts
    ```

    Override connection options:

    ``` yaml
        provisioner:
          name: ansible
          connection_options:
            ansible_ssh_user: foo
            ansible_ssh_common_args: -o IdentitiesOnly=no
    ```

    Override tags:

    ``` yaml
        provisioner:
        name: ansible
        config_options:
          tags:
            run: tag1,tag2,tag3
    ```

    A typical use case is if you want to use tags within a scenario.
    Don't forget to add a tag `always` in `converge.yml` as below.

    ``` yaml
        ---
        - name: Converge
          hosts: all
          tasks:
            - name: "Include acme.my_role_name"
              ansible.builtin.include_role:
                name: "acme.my_role_name"
              tags:
                - always
    ```

    .. _`variables defined in a playbook`: https://docs.ansible.com/ansible/latest/user_guide/playbooks_variables.html#defining-variables-in-a-playbook

    Add arguments to ansible-playbook when running converge:

    ``` yaml
        provisioner:
          name: ansible
          ansible_args:
            - --inventory=mygroups.yml
            - --limit=host1,host2
    ```
    """

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a fresh scenario logger with current context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        # Get step context from the current action being executed
        step_name = getattr(self._config, "action", "provisioner")
        return logger.get_scenario_logger(__name__, self._config.scenario.name, step_name)

    @property
    def default_config_options(self) -> dict[str, Any]:
        """Provide default options to construct ansible.cfg.

        Returns:
            Default config options.
        """
        return {
            "defaults": {
                "ansible_managed": "Ansible managed: Do NOT edit this file manually!",
                "display_failed_stderr": True,
                "forks": 50,
                "retry_files_enabled": False,
                "host_key_checking": False,
                "nocows": 1,
                "interpreter_python": "auto_silent",
            },
            "ssh_connection": {
                "scp_if_ssh": True,
                "control_path": "%(directory)s/%%h-%%p-%%r",
            },
        }

    @property
    def default_options(self) -> dict[str, str | bool]:
        """Provide default options.

        Returns:
            Default options.
        """
        d: dict[str, str | bool] = {"skip-tags": "molecule-notest,notest"}

        if self._config.action == "idempotence":
            d["skip-tags"] += ",molecule-idempotence-notest"  # type: ignore[assignment, operator]

        if self._config.debug:
            d["vvv"] = True
            d["diff"] = True

        return d

    @property
    def default_env(self) -> dict[str, str]:
        """Provide default environment variables.

        Returns:
            Default set of environment variables.
        """
        env = util.merge_dicts(
            dict(os.environ),
            {
                "ANSIBLE_CONFIG": self.config_file,
            },
        )
        env = util.merge_dicts(env, self._config.env)
        return env  # noqa: RET504

    @property
    def name(self) -> str:
        """Provisioner name.

        Returns:
            The provisioner name.
        """
        return self._config.config["provisioner"]["name"]

    @property
    def ansible_args(self) -> list[str]:
        """Provisioner ansible args.

        Returns:
            The provisioner ansible_args.
        """
        return self._config.config["provisioner"]["ansible_args"]

    @property
    def config_options(self) -> dict[str, Any]:
        """Provisioner config options.

        Returns:
            The provisioner config options.
        """
        return util.merge_dicts(
            self.default_config_options,
            self._config.config["provisioner"]["config_options"],
        )

    @property
    def options(self) -> Options:
        """Appropriate options for provisioner.

        Returns:
            Dictionary of provisioner options.
        """
        if self._config.action in ["create", "destroy"]:
            return self.default_options

        opts = self._config.config["provisioner"]["options"]
        # NOTE(retr0h): Remove verbose options added by the user while in
        # debug.
        if self._config.debug:
            opts = util.filter_verbose_permutation(opts)

        return util.merge_dicts(self.default_options, opts)

    @property
    def env(self) -> dict[str, str]:
        """Full computed environment variables for provisioner.

        Returns:
            Complete set of collected environment variables.
        """
        default_env = self.default_env
        env = self._config.config["provisioner"]["env"].copy()
        # ensure that all keys and values are strings
        env = {str(k): str(v) for k, v in env.items()}
        return util.merge_dicts(default_env, env)

    @property
    def hosts(self) -> dict[str, str]:
        """Provisioner inventory hosts.

        Returns:
            Dictionary of host names.
        """
        return self._config.config["provisioner"]["inventory"]["hosts"]

    @property
    def host_vars(self) -> dict[str, str]:
        """Provisioner inventory host vars.

        Returns:
            Dictionary of host vars.
        """
        return self._config.config["provisioner"]["inventory"]["host_vars"]

    @property
    def group_vars(self) -> dict[str, str]:
        """Provisioner inventory group vars.

        Returns:
            Dictionary of group vars.
        """
        return self._config.config["provisioner"]["inventory"]["group_vars"]

    @property
    def links(self) -> dict[str, str]:
        """Provisioner inventory links.

        Returns:
            Dictionary of links.
        """
        return self._config.config["provisioner"]["inventory"]["links"]

    @property
    def inventory(self) -> dict[str, str]:
        """Create an inventory structure and returns a dict.

        ``` yaml
            ungrouped:
              vars:
                foo: bar
              hosts:
                instance-1:
                instance-2:
              children:
                $child_group_name:
                  hosts:
                    instance-1:
                    instance-2:
            $group_name:
              hosts:
                instance-1:
                  ansible_connection: docker
                instance-2:
                  ansible_connection: docker
        ```
        """
        dd = self._vivify()
        for platform in self._config.platforms.instances:
            for group in platform.get("groups", ["ungrouped"]):
                instance_name = platform["name"]
                connection_options = self.connection_options(instance_name)
                molecule_vars = {
                    "molecule_file": "{{ lookup('env', 'MOLECULE_FILE') }}",
                    "molecule_ephemeral_directory": "{{ lookup('env', 'MOLECULE_EPHEMERAL_DIRECTORY') }}",
                    "molecule_scenario_directory": "{{ lookup('env', 'MOLECULE_SCENARIO_DIRECTORY') }}",
                    "molecule_yml": "{{ lookup('file', molecule_file) | from_yaml }}",
                    "molecule_instance_config": "{{ lookup('env', 'MOLECULE_INSTANCE_CONFIG') }}",
                    "molecule_no_log": "{{ lookup('env', 'MOLECULE_NO_LOG') or not "
                    "molecule_yml.provisioner.log|default(False) | bool }}",
                }

                # All group
                dd["all"]["hosts"][instance_name] = connection_options
                dd["all"]["vars"] = molecule_vars
                # Named group
                dd[group]["hosts"][instance_name] = connection_options
                dd[group]["vars"] = molecule_vars
                # Ungrouped
                dd["ungrouped"]["vars"] = {}
                # Children
                for child_group in platform.get("children", []):
                    dd[group]["children"][child_group]["hosts"][instance_name] = connection_options

        return self._default_to_regular(dd)

    @property
    def inventory_directory(self) -> str:
        """Inventory directory path.

        Returns:
            Path to the directory containing inventory files.
        """
        return self._config.scenario.inventory_directory

    @property
    def inventory_file(self) -> str:
        """Inventory file path.

        Returns:
            Path to ansible_inventory.yml
        """
        return str(Path(self.inventory_directory, "ansible_inventory.yml"))

    @property
    def config_file(self) -> str:
        """Configuration file path.

        Returns:
            Path to ansible.cfg.
        """
        return str(
            Path(
                self._config.scenario.ephemeral_directory,
                "ansible.cfg",
            ),
        )

    @cached_property
    def playbooks(self) -> ansible_playbooks.AnsiblePlaybooks:
        """Ansible playbooks provisioner instance.

        Returns:
            AnsiblePlaybooks instance based on this config.
        """
        return ansible_playbooks.AnsiblePlaybooks(self._config)

    @property
    def directory(self) -> str:
        """Ansible provisioner directory.

        Returns:
            Path to the ansible provisioner directory.
        """
        return str(Path(__file__).parent.parent.parent / "molecule" / "provisioner" / "ansible")

    def cleanup(self) -> None:
        """Execute `ansible-playbook` against the cleanup playbook and returns None."""
        pb = self._get_ansible_playbook(self.playbooks.cleanup)
        pb.execute()

    def connection_options(
        self,
        instance_name: str,
    ) -> dict[str, Any]:
        """Computed connection options combining instance and provisioner options.

        Args:
            instance_name: The instance to base the connection options on.

        Returns:
            The combined connection options.
        """
        d = self._config.driver.ansible_connection_options(instance_name)

        return util.merge_dicts(
            d,
            self._config.config["provisioner"]["connection_options"],
        )

    def check(self) -> None:
        """Execute ``ansible-playbook`` against the converge playbook with the ``--check`` flag."""
        pb = self._get_ansible_playbook(self.playbooks.converge)
        pb.add_cli_arg("check", value=True)
        pb.execute()

    def converge(self, playbook: str = "", **kwargs: object) -> str:
        """Execute ``ansible-playbook`` against the converge playbook. unless specified otherwise.

        Args:
            playbook: An optional string containing an absolute path to a playbook.
            **kwargs: An optional keyword arguments.

        Returns:
            str: The output from the ``ansible-playbook`` command.
        """
        pb = self._get_ansible_playbook(playbook or self.playbooks.converge, **kwargs)  # type: ignore[arg-type]

        return pb.execute()

    def destroy(self) -> None:
        """Execute ``ansible-playbook`` against the destroy playbook and returns None."""
        pb = self._get_ansible_playbook(self.playbooks.destroy)
        pb.execute()

    def side_effect(self, action_args: list[str] | None = None) -> None:
        """Execute ``ansible-playbook`` against the side_effect playbook.

        Args:
            action_args: Arguments to pass to the side_effect playbook.
        """
        playbooks = []
        if action_args:
            playbooks = [
                self._get_ansible_playbook(self.abs_path(playbook)) for playbook in action_args
            ]
        else:
            playbooks = [self._get_ansible_playbook(self.playbooks.side_effect)]
        for pb in playbooks:
            pb.execute()

    def create(self) -> None:
        """Execute ``ansible-playbook`` against the create playbook and returns None."""
        pb = self._get_ansible_playbook(self.playbooks.create)
        pb.execute()

    def prepare(self) -> None:
        """Execute ``ansible-playbook`` against the prepare playbook and returns None."""
        pb = self._get_ansible_playbook(self.playbooks.prepare)
        pb.execute()

    def syntax(self) -> None:
        """Execute `ansible-playbook` against the converge playbook with the -syntax-check flag."""
        pb = self._get_ansible_playbook(self.playbooks.converge)
        pb.add_cli_arg("syntax-check", value=True)
        pb.execute()

    def verify(self, action_args: list[str] | None = None) -> None:
        """Execute ``ansible-playbook`` against the verify playbook.

        Args:
            action_args: Arguments to pass on to the verify playbook.
        """
        playbooks = []
        if action_args:
            playbooks = [self.abs_path(playbook) for playbook in action_args]
        elif self.playbooks.verify:
            playbooks = [self.playbooks.verify]
        if not playbooks:
            self._log.warning("Skipping, verify playbook not configured.")
            return
        for playbook in playbooks:
            # Get ansible playbooks for `verify` instead of `provision`
            pb = self._get_ansible_playbook(playbook, verify=True)
            pb.execute()

    def write_config(self) -> None:
        """Write the provisioner's config file to disk and returns None."""
        template = util.render_template(
            self._get_config_template(),
            config_options=self.config_options,
        )
        util.write_file(self.config_file, template)

    def manage_inventory(self) -> None:
        """Manage inventory for Ansible and returns None."""
        self._write_inventory()
        self._remove_vars()
        if not self.links:
            self._add_or_update_vars()
        else:
            self._link_or_update_vars()

    def abs_path(self, path: str | Path) -> str:
        """Return absolute scenario-adjacent path.

        Args:
            path: Scenario-adjacent relative path.

        Returns:
            Absolute path.
        """
        path = Path(self._config.scenario.directory) / path
        return str(util.abs_path(path))

    def _add_or_update_vars(self) -> None:
        """Create host and/or group vars and returns None."""
        # Create the hosts extra inventory source (only if not empty)
        hosts_file = os.path.join(self.inventory_directory, "hosts")  # noqa: PTH118
        if self.hosts:
            util.write_file(hosts_file, util.safe_dump(self.hosts))
        # Create the host_vars and group_vars directories
        for target in ["host_vars", "group_vars"]:
            if target == "host_vars":
                vars_target = copy.deepcopy(self.host_vars)
                for instance_name in self.host_vars.keys():  # noqa: SIM118
                    instance_key = instance_name
                    vars_target[instance_key] = vars_target.pop(instance_name)

            elif target == "group_vars":
                vars_target = self.group_vars

            if vars_target:
                target_vars_directory = util.abs_path(Path(self.inventory_directory) / target)
                if not target_vars_directory.is_dir():
                    target_vars_directory.mkdir()

                for target in vars_target:  # noqa: PLW2901
                    target_var_content = vars_target[target]
                    path = target_vars_directory / target
                    util.write_file(path, util.safe_dump(target_var_content))

    def _write_inventory(self) -> None:
        """Write the provisioner's inventory file to disk and returns None."""
        self._verify_inventory()

        util.write_file(self.inventory_file, util.safe_dump(self.inventory))

    def _remove_vars(self) -> None:
        """Remove hosts/host_vars/group_vars and returns None."""
        for name in ("hosts", "group_vars", "host_vars"):
            d = os.path.join(self.inventory_directory, name)  # noqa: PTH118
            if os.path.islink(d) or os.path.isfile(d):  # noqa: PTH113, PTH114
                os.unlink(d)  # noqa: PTH108
            elif os.path.isdir(d):  # noqa: PTH112
                shutil.rmtree(d)

    def _link_or_update_vars(self) -> None:
        """Create or updates the symlink to group_vars and returns None.

        Raises:
            MoleculeError: if source file does not exist.
        """
        for d, source in self.links.items():
            target = os.path.join(self.inventory_directory, d)  # noqa: PTH118
            source = os.path.join(self._config.scenario.directory, source)  # noqa: PTH118, PLW2901

            if not os.path.exists(source):  # noqa: PTH110
                msg = f"The source path '{source}' does not exist."
                raise MoleculeError(msg)
            if os.path.exists(target):  # noqa: PTH110
                if os.path.realpath(target) == os.path.realpath(source):
                    msg = f"Required symlink {target} to {source} exist, skip creation"
                    self._log.debug(msg)
                    continue
                msg = f"Required symlink {target} exist with another source"
                self._log.debug(msg)
                os.remove(target)  # noqa: PTH107
            msg = f"Inventory {source} linked to {target}"
            self._log.debug(msg)
            os.symlink(source, target)

    def _get_ansible_playbook(
        self,
        playbook: str | None,
        verify: bool = False,  # noqa: FBT001, FBT002
        **kwargs: object,
    ) -> ansible_playbook.AnsiblePlaybook:
        """Get an instance of AnsiblePlaybook and returns it.

        Args:
            playbook: A string containing an absolute path to a provisioner's playbook.
            verify: An optional bool to toggle the Playbook mode between provision and verify.
                False: provision; True: verify. Default is False.
            **kwargs: An optional keyword arguments.

        Returns:
            An AnsiblePlaybook object.
        """
        return ansible_playbook.AnsiblePlaybook(
            playbook,
            self._config,
            verify=verify,
            **kwargs,
        )

    def _verify_inventory(self) -> None:
        """Verify the inventory is valid and returns None.

        Raises:
            MoleculeError: if inventory is missing.
        """
        if not self.inventory:
            msg = "Instances missing from the 'platform' section of molecule.yml."
            raise MoleculeError(msg)

    def _get_config_template(self) -> str:
        """Return a config template string.

        Returns:
            str
        """
        return """
{% for section, section_dict in config_options.items() -%}
[{{ section }}]
{% for k, v in section_dict.items() -%}
{{ k }} = {{ v }}
{% endfor -%}
{% endfor -%}
""".strip()

    def _vivify(self) -> Vivify:
        """Return an autovivification default dict.

        Returns:
            A defaultdict whose default value is other defaultdict objects (and so on).
        """
        return collections.defaultdict(self._vivify)

    def _default_to_regular(
        self,
        d: dict[str, Any] | collections.defaultdict[str, Any],
    ) -> dict[str, Any]:
        if isinstance(d, collections.defaultdict):
            d = {k: self._default_to_regular(v) for k, v in d.items()}

        return d

    def _get_plugin_directory(self) -> str:
        return os.path.join(self.directory, "plugins")  # noqa: PTH118

    def _absolute_path_for(self, env: dict[str, str], key: str) -> str:
        return ":".join([self.abs_path(p) for p in env[key].split(":")])



"""Custom Molecule exceptions."""

from __future__ import annotations

import logging

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Sequence
    from warnings import WarningMessage


LOG = logging.getLogger(__name__)


class MoleculeError(Exception):
    """Generic Molecule error."""

    def __init__(
        self,
        message: str = "",
        code: int = 1,
        warns: Sequence[WarningMessage] = (),
    ) -> None:
        """Custom exception to handle scenario run failures.

        Args:
            message: The message to display about the problem.
            code: Exit code to use when exiting.
            warns: Warnings about the problem to issue.
        """
        super().__init__()

        if message:
            LOG.critical(message, extra={"highlighter": False})

        for warn in warns:
            LOG.warning(warn.__dict__["message"].args[0])

        self.code = code


class ScenarioFailureError(MoleculeError):
    """Details about a scenario that failed."""



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import logging

import pytest

from molecule import config
from molecule.dependency import shell


@pytest.fixture
def _dependency_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "dependency": {
            "name": "shell",
            "command": "ls -l -a /tmp",
            "options": {"foo": "bar"},
            "env": {"FOO": "bar"},
        },
    }


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
@pytest.fixture
def _instance(  # type: ignore[no-untyped-def]  # noqa: ANN202
    _dependency_section_data,
    patched_config_validate,
    config_instance: config.Config,
):
    return shell.Shell(config_instance)


def test_shell_config_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert isinstance(_instance._config, config.Config)


def test_shell_default_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {}  # type: ignore[var-annotated]

    assert x == _instance.default_options


def test_shell_default_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert "MOLECULE_FILE" in _instance.default_env
    assert "MOLECULE_INVENTORY_FILE" in _instance.default_env
    assert "MOLECULE_SCENARIO_DIRECTORY" in _instance.default_env
    assert "MOLECULE_INSTANCE_CONFIG" in _instance.default_env


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_shell_name_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.name == "shell"


def test_shell_enabled_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.enabled


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_shell_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {"foo": "bar"}

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_shell_options_property_handles_cli_args(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.args = {}
    x = {"foo": "bar"}

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_shell_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.env["FOO"] == "bar"


def test_shell_execute(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    caplog: pytest.LogCaptureFixture,
    _instance,  # noqa: PT019
):
    # Configure caplog to capture the correct logger with scenario context
    with caplog.at_level(logging.INFO, logger="molecule.molecule.logger"):
        _instance._sh_command = "patched-command"
        _instance.execute()

        patched_run_command.assert_called_once_with(
            "patched-command",
            debug=False,
            check=True,
        )

        # Check for scenario-aware log records instead of text
        assert len(caplog.records) >= 1
        success_records = [
            r for r in caplog.records if "Dependency completed successfully" in r.getMessage()
        ]
        assert len(success_records) >= 1

        # Verify scenario context is present
        record = success_records[0]
        assert hasattr(record, "molecule_scenario")
        assert record.molecule_scenario == "default"


def test_shell_execute_does_not_execute_when_disabled(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    caplog: pytest.LogCaptureFixture,
    _instance,  # noqa: PT019
):
    _instance._config.config["dependency"]["enabled"] = False
    _instance.execute()

    assert not patched_run_command.called

    msg = "Skipping, dependency is disabled."
    assert msg in caplog.text


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_dependency_execute_bakes(patched_run_command, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance.execute()
    assert patched_run_command.call_count == 1


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_dep_executes_catches_and_exits_return_code(patched_run_command, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    patched_run_command.side_effect = SystemExit(1)
    with pytest.raises(SystemExit) as e:
        _instance.execute()
    assert e.value.code == 1


def test_has_command_configured(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._has_command_configured()



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os

from typing import TYPE_CHECKING

import pytest

from molecule import config, state, util


if TYPE_CHECKING:
    from typing import Any


@pytest.fixture
def _instance(config_instance: config.Config) -> state.State:
    return state.State(config_instance)


def test_state_file_property(_instance: state.State) -> None:  # noqa: PT019, D103
    x = os.path.join(_instance._config.scenario.ephemeral_directory, "state.yml")  # noqa: PTH118

    assert x == _instance.state_file


def test_converged(_instance: state.State) -> None:  # noqa: PT019, D103
    assert not _instance.converged


def test_created(_instance: state.State) -> None:  # noqa: PT019, D103
    assert not _instance.created


def test_state_driver(_instance: state.State) -> None:  # noqa: PT019, D103
    assert not _instance.driver


def test_prepared(_instance: state.State) -> None:  # noqa: PT019, D103
    assert not _instance.prepared


def test_reset(_instance: state.State) -> None:  # noqa: PT019, D103
    assert not _instance.converged

    _instance.change_state("converged", True)  # noqa: FBT003
    assert _instance.converged

    _instance.reset()
    assert not _instance.converged


def test_reset_persists(_instance: state.State) -> None:  # noqa: PT019, D103
    assert not _instance.converged

    _instance.change_state("converged", True)  # noqa: FBT003
    assert _instance.converged

    _instance.reset()
    assert not _instance.converged

    d = util.safe_load_file(_instance.state_file)
    assert not d.get("converged")


def test_change_state_converged(_instance: state.State) -> None:  # noqa: PT019, D103
    _instance.change_state("converged", True)  # noqa: FBT003

    assert _instance.converged


def test_change_state_created(_instance: state.State) -> None:  # noqa: PT019, D103
    _instance.change_state("created", True)  # noqa: FBT003

    assert _instance.created


def test_change_state_driver(_instance: state.State) -> None:  # noqa: PT019, D103
    _instance.change_state("driver", "foo")

    assert _instance.driver == "foo"


def test_change_state_prepared(_instance: state.State) -> None:  # noqa: PT019, D103
    _instance.change_state("prepared", True)  # noqa: FBT003

    assert _instance.prepared


def test_change_state_raises(_instance: state.State) -> None:  # noqa: PT019, D103
    with pytest.raises(state.InvalidState):
        _instance.change_state("invalid-state", True)  # noqa: FBT003


def test_get_data_loads_existing_state_file(  # noqa: D103
    _instance: state.State,  # noqa: PT019
    molecule_data: dict[str, Any],
    config_instance: config.Config,
) -> None:
    data = {"converged": False, "created": True, "driver": None, "prepared": None}
    util.write_file(_instance._state_file, util.safe_dump(data))

    s = state.State(config_instance)

    assert not s.converged
    assert s.created
    assert not s.driver
    assert not s.prepared



"""Tests for the ansi_output module."""

from __future__ import annotations

import pytest

from molecule.ansi_output import AnsiOutput, should_do_markup, to_bool


@pytest.mark.parametrize(
    ("input_value", "expected"),
    (
        (None, False),
        (True, True),
        (False, False),
        ("yes", True),
        ("YES", True),
        ("on", True),
        ("ON", True),
        ("1", True),
        ("true", True),
        ("TRUE", True),
        ("no", False),
        ("off", False),
        ("0", False),
        ("false", False),
        ("random", False),
        (1, True),
        (0, False),
        (42, False),
    ),
)
def test_to_bool(input_value: object, expected: bool) -> None:  # noqa: FBT001
    """Test to_bool function with various inputs."""
    assert to_bool(input_value) is expected


@pytest.mark.parametrize(
    ("env_vars", "expected"),
    (
        ({"NO_COLOR": "1"}, False),
        ({"FORCE_COLOR": "1"}, True),
        ({"TERM": "xterm-256color"}, True),
        ({"TERM": "dumb"}, False),
    ),
)
def test_should_do_markup(
    env_vars: dict[str, str],
    expected: bool,  # noqa: FBT001
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test should_do_markup function with various environment variables."""
    # Clear all color-related environment variables first
    for var in ["NO_COLOR", "FORCE_COLOR", "PY_COLORS", "CLICOLOR", "ANSIBLE_FORCE_COLOR", "TERM"]:
        monkeypatch.delenv(var, raising=False)

    # Set the test environment variables
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)

    assert should_do_markup() is expected


def test_ansi_output_initialization() -> None:
    """Test AnsiOutput class initialization."""
    output = AnsiOutput()
    assert hasattr(output, "markup_enabled")
    assert hasattr(output, "markup_map")
    assert isinstance(output.markup_map, dict)


def test_ansi_color_constants() -> None:
    """Test that ANSI color constants are defined."""
    output = AnsiOutput()
    assert output.RESET == "\033[0m"
    assert output.RED == "\033[31m"
    assert output.GREEN == "\033[32m"
    assert output.BLUE == "\033[34m"
    assert output.BOLD == "\033[1m"
    assert output.DIM == "\033[2m"


@pytest.mark.parametrize(
    ("input_text", "expected_output"),
    (
        ("[red]Error message[/] with [bold]bold text[/]", "Error message with bold text"),
        ("Plain text message", "Plain text message"),
        ("[info]Running [scenario]test[/] > [action]create[/][/]", "Running test > create"),
    ),
)
def test_strip_markup(input_text: str, expected_output: str) -> None:
    """Test markup stripping functionality."""
    output = AnsiOutput()
    assert output.strip_markup(input_text) == expected_output


def test_process_markup_disabled(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test markup processing when markup is disabled."""
    monkeypatch.setenv("NO_COLOR", "1")
    output = AnsiOutput()

    text_with_markup = "[red]Error message[/] with [bold]bold text[/]"
    expected = "Error message with bold text"
    assert output.process_markup(text_with_markup) == expected


def test_process_markup_enabled(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test markup processing when markup is enabled."""
    # Clear NO_COLOR and set a color environment
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("FORCE_COLOR", "1")
    output = AnsiOutput()

    text_with_markup = "[red]Error[/] message"
    result = output.process_markup(text_with_markup)

    # Should contain ANSI codes
    assert "\033[31m" in result  # Red color
    assert "\033[0m" in result  # Reset


def test_process_markup_with_unknown_tags(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test markup processing with unknown tags."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("FORCE_COLOR", "1")
    output = AnsiOutput()

    text_with_unknown = "[unknown_tag]Text[/] with [red]known tag[/]"
    result = output.process_markup(text_with_unknown)

    # Should process known tags and ignore unknown ones
    assert "\033[31m" in result  # Red color for known tag
    assert "\033[0m" in result  # Reset
    assert "Text" in result
    assert "known tag" in result


@pytest.mark.parametrize(
    ("markup_enabled", "scenario_name", "expected_pattern"),
    (
        (False, "test_scenario", "[test_scenario]"),
        (True, "test_scenario", r"\033\[32m.*\[test_scenario\].*\033\[0m"),
    ),
)
def test_format_scenario(
    markup_enabled: bool,  # noqa: FBT001
    scenario_name: str,
    expected_pattern: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test scenario formatting with markup enabled/disabled."""
    if markup_enabled:
        monkeypatch.delenv("NO_COLOR", raising=False)
        monkeypatch.setenv("FORCE_COLOR", "1")
    else:
        monkeypatch.setenv("NO_COLOR", "1")

    output = AnsiOutput()
    colored, plain = output.format_scenario(scenario_name)

    if markup_enabled:
        assert "\033[32m" in colored  # Green color for scenario tag
        assert "\033[0m" in colored  # Reset
        assert "test_scenario" in colored  # Just check for scenario name, not brackets
        assert "test_scenario" in plain  # Plain version should also contain scenario name
        assert "\033[" not in plain  # No ANSI codes in plain version
    else:
        assert colored == expected_pattern
        assert plain == expected_pattern


def test_format_scenario_with_step(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test scenario formatting with step parameter."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("FORCE_COLOR", "1")
    output = AnsiOutput()

    # Test with step
    colored, plain = output.format_scenario("test_scenario", "converge")
    assert "\033[32m" in colored  # Green for scenario
    assert "\033[33m" in colored  # Yellow for action (step)
    assert "test_scenario" in colored
    assert "converge" in colored
    assert "➜" in colored  # Right arrow
    assert ":" in colored  # Colon at end

    # Check plain version
    assert "test_scenario" in plain
    assert "converge" in plain
    assert "\033[" not in plain  # No ANSI codes in plain version

    # Test without markup
    monkeypatch.setenv("NO_COLOR", "1")
    output_no_markup = AnsiOutput()
    colored_no_markup, plain_no_markup = output_no_markup.format_scenario(
        "test_scenario",
        "converge",
    )
    assert colored_no_markup == "[test_scenario > converge]"
    assert plain_no_markup == "[test_scenario > converge]"


@pytest.mark.parametrize(
    ("level_name", "expected_ansi"),
    (
        ("DEBUG", "\033[2m"),  # Dim for DEBUG
        ("INFO", "\033[36m"),  # Cyan for INFO (new Ansible-aligned scheme)
        ("WARNING", "\033[35m"),  # Magenta for WARNING
        ("ERROR", "\033[31m"),  # Red for ERROR (new Ansible-aligned scheme)
    ),
)
def test_format_log_level_markup_enabled(
    level_name: str,
    expected_ansi: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test log level formatting when markup is enabled."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("FORCE_COLOR", "1")
    output = AnsiOutput()

    colored, plain = output.format_log_level(level_name)
    assert expected_ansi in colored
    assert "\033[0m" in colored  # Reset
    assert level_name in plain
    assert "\033[" not in plain  # No ANSI codes in plain version


def test_format_log_level_markup_disabled(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test log level formatting when markup is disabled."""
    monkeypatch.setenv("NO_COLOR", "1")
    output = AnsiOutput()

    colored, plain = output.format_log_level("INFO")
    assert colored == "INFO    "  # 8 characters, left-aligned
    assert plain == "INFO    "  # Both should be the same when markup disabled
    assert "\033[" not in colored  # No ANSI codes
    assert "\033[" not in plain  # No ANSI codes


def test_markup_map_contains_expected_styles() -> None:
    """Test that markup_map contains expected style mappings."""
    output = AnsiOutput()

    # Test basic styles from Molecule's theme
    expected_styles = [
        "info",
        "warning",
        "danger",
        "scenario",
        "action",
        "exec_phase",
        "logging.level.info",
        "logging.level.warning",
        "logging.level.error",
        "red",
        "green",
        "blue",
        "bold",
        "dim",
    ]

    for style in expected_styles:
        assert style in output.markup_map


def test_markup_map_values_are_ansi_codes() -> None:
    """Test that markup_map values are valid ANSI escape codes."""
    output = AnsiOutput()

    for ansi_code in output.markup_map.values():
        assert isinstance(ansi_code, str)
        if ansi_code:  # Some might be empty strings
            assert ansi_code.startswith("\033[")


def test_complex_markup_processing(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test processing of complex markup with nested tags."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("FORCE_COLOR", "1")
    output = AnsiOutput()

    complex_markup = "[info]Running [scenario]test[/] > [action]create[/][/]"
    result = output.process_markup(complex_markup)

    # Should contain multiple ANSI codes and resets
    assert result.count("\033[") > 1  # Multiple ANSI sequences
    assert "\033[0m" in result  # Reset codes
    assert "Running" in result
    assert "test" in result
    assert "create" in result


def test_exec_phase_markup_processing(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test processing of exec_phase markup."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("FORCE_COLOR", "1")
    output = AnsiOutput()

    # Test exec_phase markup
    result = output.process_markup("[exec_phase]Starting[/]")

    # Should contain bright cyan for exec_phase
    assert "\033[96m" in result  # Bright cyan for exec_phase
    assert "\033[0m" in result  # Reset code
    assert "Starting" in result



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import logging

from typing import TYPE_CHECKING

import pytest

from molecule.command import idempotence
from molecule.exceptions import ScenarioFailureError


if TYPE_CHECKING:
    from unittest.mock import Mock

    from pytest_mock import MockerFixture

    from molecule import config


@pytest.fixture
def _patched_is_idempotent(mocker: MockerFixture) -> Mock:
    return mocker.patch("molecule.command.idempotence.Idempotence._is_idempotent")


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
@pytest.fixture
def _instance(patched_config_validate, config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    config_instance.state.change_state("converged", True)  # noqa: FBT003

    return idempotence.Idempotence(config_instance)


def test_idempotence_execute(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    patched_ansible_converge,
    _patched_is_idempotent: Mock,  # noqa: PT019
    _instance,  # noqa: PT019
):
    with caplog.at_level(logging.INFO, logger="molecule.molecule.logger"):
        _instance.execute()

    # Check that we have log records with scenario and step information
    assert any(
        hasattr(record, "molecule_scenario")
        and hasattr(record, "molecule_step")
        and record.molecule_step == "idempotence"
        for record in caplog.records
    )

    patched_ansible_converge.assert_called_once_with()

    _patched_is_idempotent.assert_called_once_with("patched-ansible-converge-stdout")


def test_execute_raises_when_not_converged(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    caplog: pytest.LogCaptureFixture,
    patched_ansible_converge,
    _instance,  # noqa: PT019
):
    _instance._config.state.change_state("converged", False)  # noqa: FBT003
    with pytest.raises(ScenarioFailureError) as e:
        _instance.execute()

    assert e.value.code == 1

    msg = "Instances not converged.  Please converge instances first."
    assert msg in caplog.text


def test_execute_raises_when_fails_idempotence(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    patched_ansible_converge,
    _patched_is_idempotent: Mock,  # noqa: PT019
    _instance,  # noqa: PT019
):
    _patched_is_idempotent.return_value = False
    with pytest.raises(ScenarioFailureError) as e:
        _instance.execute()

    assert e.value.code == 1

    msg = "Idempotence test failed because of the following tasks:\n"
    assert msg in caplog.text


def test_is_idempotent(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    output = """
PLAY RECAP ***********************************************************
check-command-01: ok=3    changed=0    unreachable=0    failed=0
    """

    assert _instance._is_idempotent(output)


def test_is_idempotent_not_idempotent(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    output = """
PLAY RECAP ***********************************************************
check-command-01: ok=2    changed=1    unreachable=0    failed=0
check-command-02: ok=2    changed=1    unreachable=0    failed=0
    """

    assert not _instance._is_idempotent(output)


def test_non_idempotent_tasks_idempotent(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    output = """
PLAY [all] ***********************************************************

GATHERING FACTS ******************************************************
ok: [check-command-01]

TASK: [Idempotence test] *********************************************
ok: [check-command-01]

PLAY RECAP ***********************************************************
check-command-01: ok=3    changed=0    unreachable=0    failed=0
"""
    result = _instance._non_idempotent_tasks(output)

    assert result == []


def test_non_idempotent_tasks_not_idempotent(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    output = """
PLAY [all] ***********************************************************

GATHERING FACTS ******************************************************
ok: [check-command-01]
ok: [check-command-02]

TASK: [Idempotence test] *********************************************
changed: [check-command-01]
changed: [check-command-02]

PLAY RECAP ***********************************************************
check-command-01: ok=2    changed=1    unreachable=0    failed=0
check-command-02: ok=2    changed=1    unreachable=0    failed=0
"""
    result = _instance._non_idempotent_tasks(output)

    assert result == [
        "* [check-command-01] => Idempotence test",
        "* [check-command-02] => Idempotence test",
    ]



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os
import shutil
import subprocess
import warnings

from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict

import pytest

from pytest import FixtureRequest  # noqa: PT013

from molecule.command import base
from molecule.util import safe_load
from tests.conftest import mac_on_gh  # pylint:disable=C0411


if TYPE_CHECKING:
    from molecule.app import App
    from molecule.types import ScenariosResults


def run(cmd: list[str], env: dict[str, str] | None = None) -> subprocess.CompletedProcess[Any]:
    """Run a command.

    Args:
        cmd: The command to run.
        env: The environment.

    Returns:
        The result.
    """
    return subprocess.run(cmd, capture_output=True, text=True, check=False, env=env)


@cache
def has_pure_docker() -> bool:
    """Check if pure working Docker is available.

    Returns:
        True if Docker is installed, working and not using podman backend.
    """
    # https://github.com/podman-desktop/podman-desktop/issues/10745
    failure = ""
    default_docker_socket = Path("/var/run/docker.sock")
    if not shutil.which("docker"):
        failure = "Docker is not installed."
    elif default_docker_socket.exists():
        # detect if docker is NOT using podman's backend (libpod)
        result = run(
            [
                "curl",
                "--silent",
                "--unix-socket",
                "/var/run/docker.sock",
                "http:/v1/_ping",
            ],
        )
        if result.returncode != 0:
            failure = "Docker does not seem to respond to default socket."
        else:
            result = run(
                [
                    "curl",
                    "--silent",
                    "--unix-socket",
                    "/var/run/docker.sock",
                    "http:/v1/libpod/_ping",
                ],
            )
            if result.stdout == "OK":
                if shutil.which("podman-mac-helper"):
                    failure = "Docker is using podman's backend, run `sudo podman-mac-helper uninstall` to disable it."
                else:
                    failure = "Docker is using podman's backend, so we will skip running docker tests as they are high likely to fail. See https://github.com/ansible-collections/community.docker/issues/660 https://github.com/containers/podman/issues/16548"
            elif result.stdout.rstrip() not in (
                "Not Found",
                '{"message":"page not found"}',
            ):
                failure = f"Unexpected response from Docker's default socket {result.stdout}."
    else:
        result = run(["docker", "context", "inspect", "--format", "{{.Endpoints.docker.Host}}"])
        if result.returncode != 0:
            failure = f"Docker was not able to report its currently active socket for current context: {result.stderr.rstrip()}"
        else:
            docker_current_socket = result.stdout.rstrip()
            warnings.warn(
                f"Docker found but not using default socket from {default_docker_socket}, we will define DOCKER_HOST={docker_current_socket} during testing. See https://github.com/ansible-collections/community.docker/issues/660#issuecomment-2607286542",
                stacklevel=2,
            )
            os.environ["DOCKER_HOST"] = result.stdout.rstrip()
    if failure:
        warnings.warn(  # noqa: B028
            failure,
            category=RuntimeWarning,
        )
        return False
    return True


@pytest.fixture(name="scenario_to_test")
def fixture_scenario_to_test(request: FixtureRequest):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return request.param


@pytest.fixture(name="scenario_name")
def fixture_scenario_name(request: FixtureRequest):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    try:
        return request.param
    except AttributeError:
        return None


@pytest.fixture(name="driver_name")
def fixture_driver_name(request: FixtureRequest) -> str | None:  # noqa: D103
    try:
        # https://stackoverflow.com/q/65334215/99834
        return request.param  # type: ignore  # noqa: PGH003
    except AttributeError:
        return None


class ParamDefault(TypedDict):
    """Typed params."""

    argnames: tuple[str, str]
    argvalues: tuple[tuple[str, str]]
    ids: str
    indirect: tuple[str, str]


PARAMS_DEFAULT: ParamDefault = {
    "argnames": ("scenario_to_test", "scenario_name"),
    "argvalues": (("driver/delegated", "default"),),
    "ids": "0",
    "indirect": ("scenario_to_test", "scenario_name"),
}


@pytest.mark.extensive
@pytest.mark.parametrize(**PARAMS_DEFAULT)
@pytest.mark.usefixtures("with_scenario")
@pytest.mark.parametrize(
    "command",
    (
        "check",
        "cleanup",
        "converge",
        "create",
        "destroy",
        "prepare",
        "syntax",
        "side-effect",
        "test",
        "verify",
    ),
)
def test_command(
    test_ephemeral_dir_env: dict[str, str],
    scenario_name: str,
    command: str,
    app: App,
) -> None:
    """Test a scenario using command.

    Args:
        test_ephemeral_dir_env: The ephemeral directory env.
        scenario_name: The scenario name.
        command: The command to run.
    """
    cmd = ["molecule", command, "--scenario-name", scenario_name]
    result = run(cmd=cmd, env=test_ephemeral_dir_env)
    assert result.returncode == 0
    if command == "syntax":
        assert "converge.yml" in result.stdout
    else:
        assert "PLAY RECAP" in result.stdout


def test_command_report(
    test_ephemeral_dir_env: dict[str, str],
) -> None:
    """Test the output of the --report flag.

    Args:
        test_ephemeral_dir_env: The ephemeral directory env.
    """
    cmd = ["molecule", "test", "--report"]
    result = run(cmd=cmd, env=test_ephemeral_dir_env)
    assert result.returncode == 0

    scenario_result: ScenariosResults = {
        "name": "default",
        "results": [
            {
                "state": "PASSED",
                "subcommand": "destroy",
            },
            {
                "state": "PASSED",
                "subcommand": "syntax",
            },
            {
                "state": "PASSED",
                "subcommand": "create",
            },
            {
                "state": "PASSED",
                "subcommand": "converge",
            },
            {
                "state": "PASSED",
                "subcommand": "idempotence",
            },
            {
                "state": "PASSED",
                "subcommand": "destroy",
            },
        ],
    }
    report = base.generate_report([scenario_result])
    assert result.stdout.strip().endswith(report.strip())


@pytest.mark.extensive
@pytest.mark.parametrize(**PARAMS_DEFAULT)
@pytest.mark.usefixtures("with_scenario")
def test_command_idempotence(
    test_ephemeral_dir_env: dict[str, str],
    scenario_name: str,
    app: App,
) -> None:
    """Test idempotence of a scenario.

    Args:
        test_ephemeral_dir_env: The ephemeral directory env.
        scenario_name: The scenario name.
    """
    cmd = ["molecule", "create", "--scenario-name", scenario_name]
    assert run(cmd=cmd, env=test_ephemeral_dir_env).returncode == 0

    cmd = ["molecule", "converge", "--scenario-name", scenario_name]
    assert run(cmd=cmd, env=test_ephemeral_dir_env).returncode == 0

    cmd = ["molecule", "idempotence", "--scenario-name", scenario_name]
    assert run(cmd=cmd, env=test_ephemeral_dir_env).returncode == 0


@pytest.mark.parametrize(
    ("scenario_to_test", "scenario_name"),
    (("dependency", "shell"), ("dependency", "ansible-galaxy")),
)
@pytest.mark.usefixtures("with_scenario")
def test_command_dependency(
    test_ephemeral_dir_env: dict[str, str],
    scenario_name: str,
) -> None:
    """Test scenario dependency.

    Args:
        test_ephemeral_dir_env: The ephemeral directory env.
        scenario_name: The scenario name.
    """
    cmd = ["molecule", "dependency", "--scenario-name", scenario_name]
    assert run(cmd, env=test_ephemeral_dir_env).returncode == 0

    # Validate that dependency worked by running converge, which make use it
    cmd = ["molecule", "converge", "--scenario-name", scenario_name]
    assert run(cmd, env=test_ephemeral_dir_env).returncode == 0


def test_command_init_scenario(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    test_ephemeral_dir_env: dict[str, str],
) -> None:
    """Test scenario initialization.

    Manually destroy the scenario after the test because the with_scenario fixture
    is not used.

    Args:
        monkeypatch: Pytest fixture.
        tmp_path: Path to the temporary directory.
        test_ephemeral_dir_env: The ephemeral directory env.
    """
    monkeypatch.chdir(tmp_path)
    scenario_name = "test-scenario"

    cmd = ["molecule", "init", "scenario", scenario_name]
    assert run(cmd).returncode == 0
    assert (tmp_path / "molecule" / scenario_name).is_dir()

    cmd = ["molecule", "test", "--scenario-name", "test-scenario", "--all"]
    assert run(cmd=cmd, env=test_ephemeral_dir_env).returncode == 0

    cmd = ["molecule", "destroy", "--scenario-name", "test-scenario"]
    assert run(cmd=cmd, env=test_ephemeral_dir_env).returncode == 0


def test_command_list_with_format_plain(
    monkeypatch: pytest.MonkeyPatch,
    test_fixture_dir: Path,
    test_ephemeral_dir_env: dict[str, str],
    app: App,
) -> None:
    """Test list command with plain format.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
        test_ephemeral_dir_env: The ephemeral directory env.
    """
    monkeypatch.chdir(test_fixture_dir / "scenarios" / "driver" / "delegated")
    cmd = ["molecule", "list", "--format", "plain"]
    result = run(cmd=cmd, env=test_ephemeral_dir_env)
    assert result.returncode == 0
    assert result.stdout == "instance        default ansible default false   false\n"


@pytest.mark.extensive
@pytest.mark.parametrize(**PARAMS_DEFAULT)
@pytest.mark.usefixtures("with_scenario")
@pytest.mark.parametrize(("platform", "missing"), (("instance", False), ("gonzo", True)))
def test_with_missing_platform_name(
    test_ephemeral_dir_env: dict[str, str],
    scenario_name: str,
    platform: str,
    missing: bool,  # noqa: FBT001
    app: App,
) -> None:
    """Test a scenario using command with missing platform name.

    Args:
        test_ephemeral_dir_env: The ephemeral directory env.
        scenario_name: The scenario name.
        platform: The platform name.
        missing: If platform name is missing.
    """
    cmd = ["molecule", "test", "-s", scenario_name, "-p", platform]
    result = run(cmd=cmd, env=test_ephemeral_dir_env)
    assert bool(result.returncode) == missing
    if missing:
        assert "Instances missing" in result.stderr
    else:
        assert "PLAY RECAP" in result.stdout


def test_role_name_check_one(
    monkeypatch: pytest.MonkeyPatch,
    test_fixture_dir: Path,
    test_ephemeral_dir_env: dict[str, str],
    app: App,
) -> None:
    """Test role name check only warns when equal to 1.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
        test_ephemeral_dir_env: The ephemeral directory env.
    """
    scenario_dir = "delegated_invalid_role_name_with_role_name_check_equals_to_1"
    monkeypatch.chdir(test_fixture_dir / "scenarios" / "driver" / scenario_dir)
    cmd = ["molecule", "dependency"]
    result = run(cmd=cmd, env=test_ephemeral_dir_env)
    assert result.returncode == 0
    string = "molecule.delegated-test does not follow current galaxy requirements"
    assert string in result.stderr
    cmd = ["molecule", "destroy"]
    result = run(cmd=cmd, env=test_ephemeral_dir_env)
    assert result.returncode == 0


def test_sample_collection(
    monkeypatch: pytest.MonkeyPatch,
    resources_folder_path: Path,
    test_ephemeral_dir_env: dict[str, str],
) -> None:
    """Test the sample collection.

    Args:
        monkeypatch: Pytest fixture.
        resources_folder_path: Path to the resources folder.
        test_ephemeral_dir_env: The ephemeral directory env.
    """
    monkeypatch.chdir(resources_folder_path / "sample-collection")
    cmd = ["molecule", "test"]
    assert run(cmd=cmd, env=test_ephemeral_dir_env).returncode == 0


@pytest.mark.usefixtures("test_cache_path")
@pytest.mark.parametrize(("scenario_name"), (("test_w_gitignore"), ("test_wo_gitignore")))
def test_with_and_without_gitignore(
    monkeypatch: pytest.MonkeyPatch,
    scenario_name: str,
    resources_folder_path: Path,
) -> None:
    """Test with and without gitignore.

    Args:
        monkeypatch: Pytest fixture.
        scenario_name: The scenario name.
        resources_folder_path: Path to the resources folder.
    """
    if scenario_name == "test_wo_gitignore":

        def mock_return(scenario_paths: list[str]) -> list[str]:
            return scenario_paths

        monkeypatch.setattr(
            "molecule.command.base.filter_ignored_scenarios",
            mock_return,
        )

    monkeypatch.chdir(resources_folder_path)

    scenario_dir = resources_folder_path / ".extensions" / "molecule" / scenario_name
    scenario_dir.mkdir(parents=True, exist_ok=True)

    molecule_file = scenario_dir / "molecule.yml"
    molecule_file.touch()

    op = base.get_configs({}, {}, glob_str="**/molecule/*/molecule.yml")

    names = [config.scenario.name for config in op]
    if scenario_name == "test_w_gitignore":
        assert scenario_name not in names
    elif scenario_name == "test_wo_gitignore":
        assert scenario_name in names


@mac_on_gh
def test_podman(monkeypatch: pytest.MonkeyPatch, test_fixture_dir: Path, app: App) -> None:
    """Execute Podman scenario.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
    """
    monkeypatch.chdir(test_fixture_dir)
    command = ["molecule", "test", "--scenario-name", "podman"]
    assert run(command).returncode == 0


@mac_on_gh
@pytest.mark.skipif(
    not has_pure_docker(),
    reason="Bug https://github.com/ansible-collections/community.docker/issues/660",
)
def test_docker(monkeypatch: pytest.MonkeyPatch, test_fixture_dir: Path, app: App) -> None:
    """Execute Docker scenario.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
    """
    monkeypatch.chdir(test_fixture_dir)
    # This trick should force Docker to use its own backend on macOS even with Podman's Docker Extensions enabled,
    # but community.docker collection is not yet able to read it due:
    # https://github.com/ansible-collections/community.docker/issues/660
    docker_socket = Path("~/Library/Containers/com.docker.docker/Data/docker-cli.sock").expanduser()
    if docker_socket.exists():
        monkeypatch.setenv("DOCKER_SOCKET", docker_socket.as_posix())

    command = ["molecule", "test", "--scenario-name", "docker"]
    proc = subprocess.run(command, capture_output=True, text=True, check=False)
    assert proc.returncode == 0


def test_smoke(monkeypatch: pytest.MonkeyPatch, test_fixture_dir: Path, app: App) -> None:
    """Execute smoke-test scenario that should spot potentially breaking changes.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
    """
    monkeypatch.chdir(test_fixture_dir)
    command = ["molecule", "test", "--scenario-name", "smoke"]
    result = run(command)
    assert result.returncode == 0, result


def test_with_backend_as_ansible_playbook(
    monkeypatch: pytest.MonkeyPatch,
    test_fixture_dir: Path,
) -> None:
    """Execute test-scenario (smoke test) that should spot potentially breaking changes.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
    """
    monkeypatch.chdir(test_fixture_dir)
    command = ["molecule", "test", "--scenario-name", "test-scenario"]
    result = run(command)
    assert result.returncode == 0, result


@mac_on_gh
def test_with_backend_as_ansible_navigator(
    monkeypatch: pytest.MonkeyPatch,
    test_fixture_dir: Path,
) -> None:
    """Execute test-scenario-for-nav (smoke test) that should spot potentially breaking changes.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
    """
    monkeypatch.chdir(test_fixture_dir)
    command = ["molecule", "test", "--scenario-name", "test-scenario-for-nav"]
    result = run(command)
    assert result.returncode == 0, result


def test_shared_actions(
    monkeypatch: pytest.MonkeyPatch,
    test_fixture_dir: Path,
) -> None:
    """Test that shared-state properly threads scenarios.

    Args:
        monkeypatch: Pytest fixture.
        test_fixture_dir: Path to the test fixture directory.
    """
    monkeypatch.chdir(test_fixture_dir)
    cmd = [
        "molecule",
        "test",
        "--shared-state",
        "-s",
        "test-scenario",
        "-s",
        "smoke",
        "--report",
    ]
    result = run(cmd=cmd)
    assert result.returncode == 0, result
    out = result.stdout
    # If we redesign --report, this needs to be updated
    run_report = safe_load(out[out.rindex("---") :])
    assert run_report[0]["name"] == "default"
    assert run_report[1]["name"] == "test-scenario"
    assert run_report[2]["name"] == "smoke"
    assert run_report[3]["name"] == "default"



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import copy

import pytest

from molecule import config, scenario, scenarios
from molecule.console import console
from molecule.exceptions import MoleculeError
from molecule.text import chomp, strip_ansi_escape


@pytest.fixture
def _instance(config_instance: config.Config) -> scenarios.Scenarios:
    config_instance_1 = copy.deepcopy(config_instance)

    config_instance_2 = copy.deepcopy(config_instance)
    config_instance_2.config["scenario"]["name"] = "foo"

    return scenarios.Scenarios([config_instance_1, config_instance_2])


def test_configs_private_member(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    assert len(_instance._configs) == 2  # noqa: PLR2004
    assert isinstance(_instance._configs[0], config.Config)
    assert isinstance(_instance._configs[1], config.Config)


def test_scenario_name_private_member(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    assert _instance._scenario_names == []


def test_scenarios_private_member(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    assert len(_instance._scenarios) == 2  # noqa: PLR2004
    assert isinstance(_instance._scenarios[0], scenario.Scenario)
    assert isinstance(_instance._scenarios[1], scenario.Scenario)


def test_scenarios_iterator(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    s = list(_instance)

    assert s[0].name == "default"
    assert s[1].name == "foo"


def test_all_property(_instance: scenarios.Scenarios) -> None:  # noqa: PT019, D103
    result = _instance.all

    assert len(result) == 2  # noqa: PLR2004
    assert result[0].name == "default"
    assert result[1].name == "foo"


def test_all_filters_on_scenario_name_property(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    _instance._scenario_names = ["default"]

    assert len(_instance.all) == 1


def test_print_matrix(  # noqa: D103
    capsys: pytest.CaptureFixture[str],
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    with console.capture() as capture:
        _instance.print_matrix()
    result = chomp(strip_ansi_escape(capture.get()))

    matrix_out = """---
default:
  - dependency
  - cleanup
  - destroy
  - syntax
  - create
  - prepare
  - converge
  - idempotence
  - side_effect
  - verify
  - cleanup
  - destroy
foo:
  - dependency
  - cleanup
  - destroy
  - syntax
  - create
  - prepare
  - converge
  - idempotence
  - side_effect
  - verify
  - cleanup
  - destroy"""
    assert matrix_out in result


def test_verify_does_not_raise_when_found(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    _instance._scenario_names = ["default"]

    _instance._verify()


def test_verify_raises_when_scenario_not_found(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
    caplog: pytest.LogCaptureFixture,
) -> None:
    _instance._scenario_names = ["invalid"]
    with pytest.raises(MoleculeError) as e:
        _instance._verify()

    assert e.value.code == 1

    msg = "Scenario 'invalid' not found.  Exiting."
    assert msg in caplog.text


def test_verify_raises_when_multiple_scenarios_not_found(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
    caplog: pytest.LogCaptureFixture,
) -> None:
    _instance._scenario_names = ["invalid", "also invalid"]
    with pytest.raises(MoleculeError) as e:
        _instance._verify()

    assert e.value.code == 1

    msg = "Scenarios 'also invalid, invalid' not found.  Exiting."
    assert msg in caplog.text


def test_filter_for_scenario(  # noqa: D103
    _instance: scenarios.Scenarios,  # noqa: PT019
) -> None:
    _instance._scenario_names = ["default"]
    result = _instance._filter_for_scenario()
    assert len(result) == 1
    assert result[0].name == "default"

    _instance._scenario_names = ["invalid"]
    result = _instance._filter_for_scenario()
    assert result == []


def test_get_matrix(_instance: scenarios.Scenarios) -> None:  # noqa: PT019, D103
    matrix = {
        "default": {
            "idempotence": ["idempotence"],
            "syntax": ["syntax"],
            "converge": ["dependency", "create", "prepare", "converge"],
            "cleanup": ["cleanup"],
            "check": [
                "dependency",
                "cleanup",
                "destroy",
                "create",
                "prepare",
                "converge",
                "check",
                "cleanup",
                "destroy",
            ],
            "verify": ["verify"],
            "create": ["dependency", "create", "prepare"],
            "prepare": ["prepare"],
            "side_effect": ["side_effect"],
            "dependency": ["dependency"],
            "test": [
                "dependency",
                "cleanup",
                "destroy",
                "syntax",
                "create",
                "prepare",
                "converge",
                "idempotence",
                "side_effect",
                "verify",
                "cleanup",
                "destroy",
            ],
            "destroy": ["dependency", "cleanup", "destroy"],
        },
        "foo": {
            "idempotence": ["idempotence"],
            "syntax": ["syntax"],
            "converge": ["dependency", "create", "prepare", "converge"],
            "check": [
                "dependency",
                "cleanup",
                "destroy",
                "create",
                "prepare",
                "converge",
                "check",
                "cleanup",
                "destroy",
            ],
            "cleanup": ["cleanup"],
            "create": ["dependency", "create", "prepare"],
            "verify": ["verify"],
            "prepare": ["prepare"],
            "side_effect": ["side_effect"],
            "dependency": ["dependency"],
            "test": [
                "dependency",
                "cleanup",
                "destroy",
                "syntax",
                "create",
                "prepare",
                "converge",
                "idempotence",
                "side_effect",
                "verify",
                "cleanup",
                "destroy",
            ],
            "destroy": ["dependency", "cleanup", "destroy"],
        },
    }

    assert matrix == _instance._get_matrix()



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os
import platform
import shutil
import subprocess

from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

from molecule.app import App, get_app


if TYPE_CHECKING:
    from collections.abc import Iterable

    from _pytest.nodes import Item
    from _pytest.runner import CallInfo

    from molecule.scenario import Scenario


FIXTURES_DIR = Path(__file__).parent / "fixtures"


# Marker to skip tests on macos when running on GH
mac_on_gh = pytest.mark.skipif(
    platform.system() == "Darwin" and "CI" in os.environ,
    reason="Podman and docker not currently available for macOS on Github",
)


def is_subset(subset, superset):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    # Checks if first dict is a subset of the second one
    if isinstance(subset, dict):
        return all(
            key in superset and is_subset(val, superset[key])  # type: ignore[no-untyped-call]
            for key, val in subset.items()
        )

    if isinstance(subset, list | set):
        return all(
            any(is_subset(subitem, superitem) for superitem in superset)  # type: ignore[no-untyped-call]
            for subitem in subset
        )

    # assume that subset is a plain value if none of the above match
    return subset == superset


@pytest.fixture(scope="session")
def app(name: str = "app") -> App:
    """Run the container.

    Returns:
        callable: The container executor.
    """
    return get_app(Path())


@pytest.fixture(autouse=True)
def _no_color(monkeypatch: pytest.MonkeyPatch) -> None:
    """Disable coloring output.

    Args:
        monkeypatch: pytest.MonkeyPatch.
    """
    # Analyzing output with no color is much easier. Tests that need to test for
    # color output, should override the value.
    monkeypatch.setitem(os.environ, "NO_COLOR", "1")
    monkeypatch.delitem(os.environ, "PY_COLORS", raising=False)
    monkeypatch.delitem(os.environ, "ANSIBLE_FORCE_COLOR", raising=False)
    monkeypatch.delitem(os.environ, "FORCE_COLOR", raising=False)


@pytest.fixture
def resources_folder_path() -> Path:
    """Return the path to the resources folder.

    Returns:
        Path: The path to the resources folder.
    """
    return FIXTURES_DIR / "resources"


def pytest_collection_modifyitems(items, config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    marker = config.getoption("-m")
    is_sharded = False
    shard_id = 0
    shards_num = 0
    if not marker.startswith("shard_"):
        return
    shard_id, _, shards_num = marker[6:].partition("_of_")
    if shards_num:
        shard_id = int(shard_id)
        shards_num = int(shards_num)
        is_sharded = True
    else:
        raise ValueError("shard_{}_of_{} marker is invalid")  # noqa: EM101, TRY003
    if not is_sharded:
        return
    if not 0 < shard_id <= shards_num:
        raise ValueError(  # noqa: TRY003
            "shard_id must be greater than 0 and not bigger than shards_num",  # noqa: EM101
        )
    for test_counter, item in enumerate(items):
        cur_shard_id = test_counter % shards_num + 1
        marker = getattr(pytest.mark, f"shard_{cur_shard_id}_of_{shards_num}")
        item.add_marker(marker)
    del marker
    print(f"Running sharded test group #{shard_id} out of {shards_num}")


@pytest.fixture(autouse=True)
def reset_pytest_vars(monkeypatch):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Make PYTEST_* env vars inaccessible to subprocesses."""
    for var_name in tuple(os.environ):
        if var_name.startswith("PYTEST_"):
            monkeypatch.delenv(var_name, raising=False)


@pytest.fixture
def test_fixture_dir(request: pytest.FixtureRequest) -> Path:
    """Provide the fixture directory for a given test.

    :param request: The pytest request object
    :returns: The fixture directory
    """
    return FIXTURES_DIR / request.path.relative_to(Path(__file__).parent).with_suffix(
        "",
    )


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_runtest_makereport(  # type: ignore[misc]
    item: Item,
    call: CallInfo[None],
) -> pytest.TestReport:
    """Create a phase report for each test phase.

    Allows the test success or failure to be available in a fixture
    https://docs.pytest.org/en/latest/example/simple.html#making-test-result-information-available-in-fixtures

    Args:
        item: The pytest item
        call: The pytest call
    Returns:
        The pytest report
    """
    outcome = yield
    rep = outcome.get_result()
    setattr(item, "rep_" + rep.when, rep)
    return rep


def git_root() -> Path:
    """Return the root of the git repository.

    Returns:
        The root of the git repository
    """
    command = "git rev-parse --show-toplevel"
    return Path(os.popen(command, mode="r").read().strip())  # noqa: S605


@pytest.fixture(name="test_cache_path")
def fixture_test_cache_path(
    monkeypatch: pytest.MonkeyPatch,
    request: pytest.FixtureRequest,
) -> Iterable[Path]:
    """Provide a test specific cache path.

    The assumption here is that the tests are always run after being checkout out from git.
    The ephemeral directory function will be mocked to ensure that the cache directory is
    used by the test. as the ephemeral directory.

    After success completion of the test, the directory will be removed.

    Args:
        monkeypatch: The pytest monkeypatch object
        request: The pytest request object

    Yields:
        Path: The ephemeral directory
    """
    test_dir = Path(
        git_root()
        / ".cache"
        / ".molecule"
        / "tests"
        / request.path.relative_to(Path(__file__).parent).with_suffix("")
        / request.node.name,
    )
    if test_dir.exists():
        shutil.rmtree(test_dir, ignore_errors=True)

    test_dir.mkdir(parents=True, exist_ok=True)

    def mock_ephemeral_directory(_self: Scenario) -> str:
        """Mock the ephemeral directory.

        Returns:
            Path: The ephemeral directory
        """
        return str(test_dir)

    monkeypatch.setattr(
        "molecule.scenario.Scenario.ephemeral_directory",
        property(mock_ephemeral_directory),
    )

    yield test_dir

    item = request.node
    if item.rep_call.failed:
        # if the test failed, keep the ephemeral directory for debugging
        return

    shutil.rmtree(test_dir, ignore_errors=True)


@pytest.fixture(name="test_ephemeral_dir_env")
def fixture_test_ephemeral_dir_env(test_cache_path: Path) -> dict[str, str]:
    """Provide the ephemeral directory for a given test as an env variable.

    The current path will be provided as an env variable for any subprocess commands to use
    so any commands can be found.

    Args:
        test_cache_path: The ephemeral directory path

    Returns:
        The ephemeral directory as an env variable
    """
    path = os.environ.get("PATH", "")
    return {
        "PATH": path,
        "MOLECULE_EPHEMERAL_DIRECTORY": str(test_cache_path),
    }


def pytest_sessionfinish(session: pytest.Session, exitstatus: int) -> None:
    """Remove the test cache directory after a successful test session.

    Args:
        session: The pytest session
        exitstatus: The exit status of the test session
    """
    assert session
    if exitstatus:
        return
    if os.environ.get("PYTEST_XDIST_WORKER"):
        return
    test_cache_dir = Path(git_root() / ".cache" / ".molecule" / "tests")
    if test_cache_dir.exists():
        shutil.rmtree(test_cache_dir)


def run(cmd: list[str], env: dict[str, str] | None = None) -> subprocess.CompletedProcess[Any]:
    """Run a command.

    Args:
        cmd: The command to run.
        env: The environment.

    Returns:
        The result.
    """
    return subprocess.run(cmd, capture_output=True, text=True, check=False, env=env)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Unittest for v2 config format."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from molecule import util


if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any


@pytest.fixture(name="molecule_file")
def fixture_molecule_file(resources_folder_path: Path) -> Path:
    """Return path to molecule file.

    Args:
        resources_folder_path: Path to the resources folder.

    Returns:
        Path to molecule file.
    """
    return resources_folder_path / "molecule.yml"


@pytest.fixture
def config(molecule_file: Path, request: pytest.FixtureRequest) -> dict[str, Any]:
    """Return merged molecule file data.

    Args:
        molecule_file: Path to molecule file.
        request: Pytest fixture request.

    Returns:
        Merged molecule file data.
    """
    with molecule_file.open() as f:
        d = util.safe_load(f)
    if hasattr(request, "param"):
        if isinstance(request.getfixturevalue(request.param), str):
            d2 = util.safe_load(request.getfixturevalue(request.param))
        else:
            d2 = request.getfixturevalue(request.param)
        d = util.merge_dicts(d, d2)

    return d  # type: ignore[no-any-return]


@pytest.fixture
def _model_platforms_delegated_section_data() -> str:
    return """
---
platforms:
  - name: instance
""".strip()



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""List Command Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from rich import box
from rich.syntax import Syntax
from rich.table import Table

from molecule import scenarios, text, util
from molecule.click_cfg import click_command_ex, options
from molecule.command import base
from molecule.console import console
from molecule.constants import MOLECULE_GLOB
from molecule.status import Status


if TYPE_CHECKING:
    import click

    from molecule.types import CommandArgs


class List(base.Base):
    """List command shows information about current scenarios."""

    def execute(  # type: ignore[override]
        self,
        action_args: list[str] | None = None,  # noqa: ARG002
    ) -> list[Status]:
        """Execute the actions necessary to perform a `molecule list`.

        Args:
            action_args: Arguments for this command. Unused.

        Returns:
            List of statuses.
        """
        return self._config.driver.status()


@click_command_ex(name="list")
@options(["scenario_name_single", "format_full"])
def list_(ctx: click.Context) -> None:  # pragma: no cover
    """List status of instances.

    Args:
        ctx: Click context object holding commandline arguments.
    """
    args = ctx.obj.get("args")
    subcommand = base._get_subcommand(__name__)  # noqa: SLF001
    command_args: CommandArgs = {
        "format": ctx.params["format"],
        "subcommand": subcommand,
    }

    output_format = ctx.params["format"]
    scenario_name = ctx.params["scenario_name"]

    statuses = []
    s = scenarios.Scenarios(
        base.get_configs(args, command_args, glob_str=MOLECULE_GLOB),
        None if scenario_name is None else [scenario_name],
    )
    for scenario in s:
        statuses.extend(base.execute_subcommand(scenario.config, subcommand))

    headers = [text.title(name) for name in Status._fields]
    if output_format in ["simple", "plain"]:
        table_format = output_format

        if output_format == "plain":
            headers = []
        _print_tabulate_data(headers, statuses, table_format)
    else:
        _print_yaml_data(headers, statuses)


def _print_tabulate_data(
    headers: list[str],
    data: list[Status],
    table_format: str,
) -> None:  # pragma: no cover
    """Show the tabulate data on the screen.

    Args:
        headers: A list of column headers.
        data: A list of tabular data to display.
        table_format: A string containing the table format.
    """
    if table_format == "plain":
        for line in data:
            console.print("\t".join(line))
    else:
        t = Table(box=box.MINIMAL)
        for header in headers:
            t.add_column(header)
        for line in data:
            t.add_row(*line)
        console.print(t)


def _print_yaml_data(
    headers: list[str],
    data: list[Status],
) -> None:  # pragma: no cover
    """Show the tabulate data on the screen in yaml format.

    Args:
        headers: A list of column headers.
        data: A list of tabular data to display.
    """
    l = [  # noqa: E741
        dict(
            zip(
                headers,
                [getattr(datum, field) for field in datum._fields],
                strict=False,
            ),
        )
        for datum in data
    ]

    syntax = Syntax(util.safe_dump(l), "yaml")
    console.print(syntax)



"""Ansible Molecule Filters."""

from __future__ import annotations

import os

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from collections.abc import MutableMapping


# pylint: skip-file

DOCUMENTATION = """
name: from_yaml
description:
    - This callback just adds total play duration to the play stats.
"""
try:
    from molecule import config, interpolation, util

    MOLECULE_IMPORT_ERROR = None
except ImportError as imp_exc:
    MOLECULE_IMPORT_ERROR = imp_exc


def from_yaml(data: Any) -> Any:  # noqa: ANN401
    """Interpolate the provided data and return a dict.

    Currently, this is used to re-interpolate the `molecule.yml` inside an
    Ansible playbook.  If there were any interpolation errors, they would
    have been found and raised earlier.

    :return: dict
    """
    if MOLECULE_IMPORT_ERROR:
        return None

    molecule_env_file = os.environ.get("MOLECULE_ENV_FILE", None)

    env: MutableMapping[str, str] = os.environ.copy()
    if molecule_env_file:
        env = config.set_env_from_file(env, molecule_env_file)  # pylint: disable=undefined-variable

    i = interpolation.Interpolator(interpolation.TemplateWithDefaults, env)  # pylint: disable=undefined-variable
    interpolated_data = i.interpolate(data)

    return util.safe_load(interpolated_data)  # pylint: disable=undefined-variable


def to_yaml(data: Any) -> str:  # noqa: ANN401
    """Format data as YAML."""
    if MOLECULE_IMPORT_ERROR:
        return ""

    return util.safe_dump(data)  # pylint: disable=undefined-variable


def header(content: str) -> str:
    """Prepend molecule header."""
    if MOLECULE_IMPORT_ERROR:
        return ""

    return util.molecule_prepender(content)


def get_docker_networks(  # noqa: C901
    data: list[dict[str, Any]],
    state: str,
    labels: dict[str, str] | None = None,
) -> list[dict[str, Any]]:
    """Get list of docker networks."""
    network_list = []
    network_names = []
    if not labels:
        labels = {}
    for platform in data:
        if "docker_networks" in platform:
            for docker_network in platform["docker_networks"]:
                if "labels" not in docker_network:
                    docker_network["labels"] = {}
                for key in labels:
                    docker_network["labels"][key] = labels[key]

                docker_network["state"] = state

                if "name" in docker_network:
                    network_list.append(docker_network)
                    network_names.append(docker_network["name"])

        # If a network name is defined for a platform but is not defined in
        # docker_networks, add it to the network list.
        if "networks" in platform:
            for network in platform["networks"]:
                if "name" in network:
                    name = network["name"]
                    if name not in network_names:
                        network_list.append(
                            {
                                "name": name,
                                "labels": labels,
                                "state": state,
                            },
                        )
    return network_list


class FilterModule:
    """Core Molecule filter plugins."""

    def filters(self) -> dict[str, Any]:
        """Return implemented filters."""
        return {
            "from_yaml": from_yaml,
            "to_yaml": to_yaml,
            "header": header,
            "get_docker_networks": get_docker_networks,
        }



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Create Command Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molecule import api
from molecule.click_cfg import click_command_ex, options
from molecule.console import console


if TYPE_CHECKING:
    import click


@click_command_ex()
@options(["format_simple"])
def drivers(ctx: click.Context) -> None:  # pragma: no cover
    """List drivers.

    Args:
        ctx: Click context object holding commandline arguments.
    """
    output_format = ctx.params["format"]
    drivers = []  # pylint: disable=redefined-outer-name

    for driver in api.drivers().values():
        description = str(driver)
        if output_format == "plain":
            description = f"{driver!s:16s}[logging.level.notset] {driver.title} Version {driver.version} from {driver.module} python module.)[/logging.level.notset]"
        drivers.append([driver, description])
        console.print(description)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Schema v3 Validation Module."""

from __future__ import annotations

import json
import logging

from pathlib import Path
from typing import TYPE_CHECKING

from jsonschema import validate as jsonschema_validate
from jsonschema.exceptions import ValidationError

from molecule import api
from molecule.data import __file__ as data_module


if TYPE_CHECKING:
    from molecule.types import ConfigData


LOG = logging.getLogger(__name__)


def validate(c: ConfigData) -> list[str]:
    """Perform schema validation.

    Args:
        c: Dictionary of configuration data.

    Returns:
        Any errors generated by the schema validation.
    """
    result = []
    schemas = []

    schema_files = [str(Path(data_module).parent / "molecule.json")]
    driver_name = c["driver"]["name"]

    driver_schema_file = None
    if driver_name in api.drivers():
        driver_schema_file = api.drivers()[driver_name].schema_file()

    if driver_schema_file is None:
        msg = f"Driver {driver_name} does not provide a schema."
        LOG.warning(msg)
    elif not Path(driver_schema_file).exists():
        msg = f"Schema {driver_schema_file} for driver {driver_name} not found."
        LOG.warning(msg)
    else:
        schema_files.append(driver_schema_file)

    for schema_file in schema_files:
        with Path(schema_file).open(encoding="utf-8") as f:
            schema = json.load(f)
        schemas.append(schema)

    try:
        for schema in schemas:
            jsonschema_validate(c, schema)
    except ValidationError as exc:
        # handle validation error for driver name
        if exc.json_path == "$.driver.name" and exc.message.endswith(
            (
                "is not of type 'string'",
                "is not valid under any of the given schemas",
            ),
        ):
            wrong_driver_name = str(exc.message.split()[0])
            if isinstance(exc.schema, dict):
                driver_name_err_msg = exc.schema["messages"]["anyOf"]
            else:
                driver_name_err_msg = "is not a valid driver name"
            result.append(f"{wrong_driver_name} {driver_name_err_msg}")
        else:
            result.append(exc.message)

    return result



"""Text utils."""

from __future__ import annotations

import base64
import hashlib
import re

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pathlib import Path
    from typing import AnyStr


def camelize(string: str) -> str:
    """Format string as camel-case.

    Args:
        string: The string to be formatted.

    Returns:
        The formatted string.
    """
    # NOTE(retr0h): Taken from jpvanhal/inflection
    # https://github.com/jpvanhal/inflection
    return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), string)


def underscore(string: str) -> str:
    """Format string to underlined notation.

    Args:
        string: The string to be formatted.

    Returns:
        The formatted string.
    """
    # NOTE(retr0h): Taken from jpvanhal/inflection
    # https://github.com/jpvanhal/inflection
    string = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", string)
    string = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", string)
    string = string.replace("-", "_")

    return string.lower()


def title(word: str) -> str:
    """Format title.

    Args:
        word: The string to be formatted.

    Returns:
        The formatted string.
    """
    return " ".join(x.capitalize() or "_" for x in word.split("_"))


def chomp(text: str) -> str:
    """Remove any training spaces from a multiline string.

    Args:
        text: Block of text to strip.

    Returns:
        Text with trailing spaces stripped from newlines.
    """
    return "\n".join([x.rstrip() for x in text.splitlines()])


def strip_ansi_escape(data: AnyStr) -> str:
    """Remove all ANSI escapes from string or bytes.

    Args:
        data: Text to clean in string or bytes.

    Returns:
        String cleaned of ANSI escape sequences.
    """
    text = _to_unicode(data)
    return re.sub(r"\x1b[^m]*m", "", text)


def strip_ansi_color(data: AnyStr) -> str:
    """Remove ANSI colors from string or bytes.

    Args:
        data: Text to clean in string or bytes.

    Returns:
        String cleaned of ANSI colors.
    """
    # Taken from tabulate
    invisible_codes = re.compile(r"\x1b\[\d*m")

    text = _to_unicode(data)
    return re.sub(invisible_codes, "", text)


def _to_unicode(data: AnyStr) -> str:
    """Ensure data is a UTF-8 string.

    Args:
        data: A string that can be bytes or str.

    Returns:
        A UTF-8 string.
    """
    if isinstance(data, bytes):
        return data.decode("utf-8")
    return data


def checksum(data: str | Path, length: int = 5) -> str:
    """Returns a checksum for the given data.

    Args:
        data: The data to checksum.
        length: The length of the checksum.

    Returns:
        A checksum string.
    """
    data = str(data)
    # Hash the input string using SHA-256
    hash_object = hashlib.sha256(data.encode("utf-8"))
    # Convert the hash to a base64-encoded string
    base64_hash = base64.urlsafe_b64encode(hash_object.digest()).decode("utf-8")
    # Truncate the result to the desired length
    return base64_hash[:length]
